import numpy as np
import pandas as pd
import math

from hypnospy import Wearable, Experiment
from scipy.stats import entropy
from scipy import linalg  # linear algebra (matrix) processing package
from tqdm import tqdm, trange
from collections import defaultdict
from tensorflow.keras.preprocessing import timeseries_dataset_from_array


class CircadianAnalysis(object):

    def __init__(self, input: {Wearable, Experiment}):
        if type(input) is Wearable:
            self.wearables = [input]
        elif type(input) is Experiment:
            self.wearables = input.get_all_wearables()

    def run_SSA(self):
        for idx, w in tqdm(enumerate(self.wearables)):
            # print(wearable.get_pid())

            if w.data.shape[0] == 0:
                wearable.data[output_col] = np.nan
                warnings.warn("No data for PID %s. Skipping it." % wearable.get_pid())
                continue

            print(idx)
            w = self._get_SSA(w)
            self.wearables[idx] = w

    # Extract SSA parameters, partial variances stored in ssa_io_pv, other stored in dict ssa, read from files in folder for
    # convenience, as full analysis takes a long time for each subject
    def _get_SSA(self, w, cols=['hyp_act_x'], freqs=['15T']):
        ssa = defaultdict(dict)
        for col in cols:
            df = w.data[['hyp_exp_day', 'hyp_time_col', 'hyp_act_x']].resample('1T', on='hyp_time_col').mean()
            df = df['hyp_act_x']
            ssa[col]['r'], ssa[col]['pv'], ssa[col]['gk'], ssa[col]['wm'] = self._get_SSA_par(df, L=1440)
            ssa[col]['df'] = df

            pd.DataFrame(ssa[col]['gk'][:10, :]).to_csv('ssa_gk_' + 'self.filename' + col + '.csv', index=False)
            pd.DataFrame(ssa[col]['wm']).to_csv('ssa_wm_' + 'self.filename' + col + '.csv', index=False)
            pd.DataFrame(ssa[col]['pv']).to_csv('ssa_pv_' + 'self.filename' + col + '.csv', index=False)
            for freq in freqs:
                df_1 = pd.DataFrame(np.transpose(ssa[col]['gk']))

                # if self.data freq is different than df.resample, then the below line will cause an error
                # df_1 = df_1.set_index(self.data.index)
                df = w.data[['hyp_exp_day', 'hyp_time_col', 'hyp_act_x']].resample('1T', on='hyp_time_col').mean()
                df = df['hyp_act_x']
                ssa[col]['df'] = df

                # df_1 = df_1.set_index(df.index)
                df_1 = df_1.set_index(ssa[col]['df'].index)

                df_2 = df_1.between_time('06:00', '23:00', include_start=True, include_end=True)
                df_gk = df_2[0] + df_2[1]
                # Get SSA acrophases
                ssa[col]['acrophase'] = df_gk.resample('24H').agg(lambda x: np.nan if x.count() == 0 else x.idxmax())
                # Get vectors of coarse-grain params for ML
                ssa[col]['gksum' + freq] = df_gk.resample(freq).mean()
                # Get trend i.e. mesor
                ssa[col]['trend'] = df_1[0].resample('24H').mean()
                # Get period in minutes
                ssa[col]['period'] = (ssa[col]['acrophase'] - ssa[col]['acrophase'].shift(1)).astype('timedelta64[m]')
        w.ssa = ssa
        return w

    @staticmethod
    def _get_SSA_par(df, L=1440):  # 2 <= L <= N/2
        N = len(df)
        K = N - L + 1

        dataset = timeseries_dataset_from_array(
            data=df,
            targets=None,
            sequence_length=L,
            sequence_stride=1,
            sampling_rate=1,
            batch_size=len(df)
        )

        X = list(dataset.as_numpy_iterator())[0]
        print(X.shape)

        try:
            U, s, V = linalg.svd(X,
                                 full_matrices=True,
                                 compute_uv=True,
                                 overwrite_a=False,
                                 check_finite=True,
                                 lapack_driver='gesvd')
        except:
            U, s, V = linalg.svd(X,
                                 full_matrices=True,
                                 compute_uv=True,
                                 overwrite_a=False,
                                 check_finite=True,
                                 lapack_driver='gesvd')

        l = s ** 2  # partial variances
        r = len(s)  # np.linalg.matrix_rank(X) # matrix rank and total number of components
        ### time-series components ###
        gkList = np.zeros(shape=(r, N))  # zero matrix in whose rows SSA components will be saved

        print('input:', X.shape)
        print('U:', U.shape)
        print('s:', s.shape)
        print('V:', V.shape)

        print('r:', r)
        print('gkList:', gkList.shape)

        for k in trange(r, position=0, leave=True):
            Uk = U[:, k]  # k-th order column singular vector
            Vk = V[k, :]  # k-th order row singular vector
            Xk = s[k] * np.outer(Uk, Vk)  # k-th order matrix component
            gk = []  # empty array in which to save successive k-th order component values
            for i in range(min(K - 1, L - 1), -max(K - 1, L - 1) - 1, -1):  # loop over diagonals
                gki = np.mean(np.diag(np.fliplr(Xk), i))  # successive time.series values
                gk.append(gki)
            gkList[k] = gk  # k-th order component

        ### w-corr matrix ###
        w = []  # empty array to which to add successive weights
        LL = min(L, K)
        KK = max(L, K)
        for ll in range(1, LL + 1):  # first 1/3 part of weights
            w.append(ll)
        for ll in range(LL + 1, KK + 1):  # second 1/3 part of weights
            w.append(LL)
        for ll in range(KK + 1, N + 1):  # third 1/3 part of weights
            w.append(N - ll)
        kMin = kkMin = 0  # show w-corr matrix for first 20 index values
        kMax = kkMax = 20

        wMatrix = [[sum(w * gkList[k] * gkList[kk]) / (
                    math.sqrt(sum(w * gkList[k] * gkList[k])) * math.sqrt(sum(w * gkList[kk] * gkList[kk]))) for k in
                    range(kMin, kMax)] for kk in range(kkMin, kkMax)]
        wMatrix = np.array(wMatrix)
        return (r, l, gkList, wMatrix);

    def run_cosinor():
        # apply cosinor method
        pass

    def run_PSD():
        # apply power spectral density analysis

        df['PSD'] = PSD
        pass

    def run_entropy(df, bins=20):
        bins = int(bins)
        hist, bin_edges = np.histogram(df, bins=bins)
        p = hist / float(hist.sum())
        ent = entropy(p)
        df['ent'] = ent
        return ent

    ## Simpler Implementation perhaps?
    def run_LIDS(timestamp, ENMO):
        """
        [1] Winnebeck, E. C., Fischer, D., Leise, T., & Roenneberg, T. (2018).
        Dynamics and Ultradian Structure of Human Sleep in Real Life.
        Current Biology, 28(1), 49–59.e5. http://doi.org/10.1016/j.cub.2017.11.063
        """

        # Read data 
        # df = pd.concat((timestamp, pd.Series(ENMO)), axis=1)
        # df.columns = ['timestamp','ENMO']
        # df.set_index('timestamp', inplace=True)

        # Set threshold
        df['ENMO_thresh'] = np.where(ENMO < 0.02, 0, ENMO - 0.02)  # assuming ENMO is in g
        # 10-minute rolling sum
        ENMO_sub_smooth = df['ENMO_thresh'].rolling('10m').sum()
        df['LIDS_unfiltered'] = 100.0 / (ENMO_sub_smooth + 1.0)
        # 30-minute rolling average
        LIDS = df['LIDS_unfiltered'].rolling('30m').mean().values
        return LIDS

    ## FOUND THE FOLLOWING FUNCTION ONLINE FROM ANOTHER PAPER (Really intersting but potentially an overkill)
    ## Will delete after as its not mine, good for now to check the structure  


class LIDS():
    """
    Class for Locomotor inactivity during sleep (LIDS) Analysis
    Winnebeck, E. C., Fischer, D., Leise, T., & Roenneberg, T. (2018).
    Dynamics and Ultradian Structure of Human Sleep in Real Life.
    Current Biology, 28(1), 49–59.e5. http://doi.org/10.1016/j.cub.2017.11.063
    """

    lids_func_list = ['lids']
    fit_func_list = ['cosine', 'chirp', 'modchirp']

    def __init__(
            self,
            lids_func='lids',
            fit_func='cosine',
            fit_obj_func='residuals',
            fit_params=None
    ):

        # LIDS functions
        lids_funcs = {'lids': _lids_func}
        if lids_func not in lids_funcs.keys():
            raise ValueError(
                '`LIDS function` must be "%s". You passed: "%s"' %
                ('" or "'.join(list(lids_funcs.keys())), lids_func)
            )

        # Fit functions
        fit_funcs = {'cosine': _cosine, 'chirp': _lfm, 'modchirp': _lfam}
        if fit_func not in fit_funcs.keys():
            raise ValueError(
                '`Fit function` must be "%s". You passed: "%s"' %
                ('" or "'.join(list(fit_funcs.keys())), fit_func)
            )

        # Fit objective functions (i.e. funcitons to be minimized)
        fit_obj_funcs = {
            'residuals': _residual,
            'nll': _lids_likelihood
        }
        # and associated functions to convert the residuals to a scalar value
        fit_reduc_funcs = {
            'residuals': None,
            'nll': _nlog
        }

        if fit_obj_func not in fit_obj_funcs.keys():
            raise ValueError(
                '`Fit objective function` must be "%s". You passed: "%s"' %
                ('" or "'.join(list(fit_obj_funcs.keys())), fit_obj_func)
            )

        self.__freq = None  # pd.Timedelta
        self.__lids_func = lids_funcs[lids_func]  # LIDS transformation fct
        self.__fit_func = fit_funcs[fit_func]  # Fit function to LIDS
        self.__fit_obj_func = fit_obj_funcs[fit_obj_func]  # Fit obj function
        self.__fit_reduc_func = fit_reduc_funcs[fit_obj_func]

        if fit_params is None:
            fit_params = Parameters()
            # Default parameters for the cosine fit function
            fit_params.add('amp', value=50, min=0, max=100)
            fit_params.add('phase', value=0.0, min=-2 * np.pi, max=2 * np.pi)
            fit_params.add('period', value=9, min=0)  # Dummy value
            # Introduce inequality amp+offset < 100
            fit_params.add('delta', value=60, max=100, vary=True)
            fit_params.add('offset', expr='delta-amp')
            # Additional parameters for the chirp fit function
            if fit_func == 'chirp':
                fit_params.add('k', value=-.0001, min=-1, max=1)
                fit_params.add('slope', value=-0.5)
            # Additional parameters for the modchirp fit function
            if fit_func == 'modchirp':
                fit_params.add('k', value=-.0001, min=-1, max=1)
                fit_params.add('slope', value=-0.5)
                fit_params.add('mod', value=0.0001, min=-10, max=10)

        self.__fit_initial_params = fit_params
        # self.__fit_params = None
        self.__fit_results = None
        # self.__fit_period = None

    @property
    def freq(self):
        r'''Sampling frequency of the LIDS transformed data'''
        if self.__freq is None:
            warnings.warn(
                'The sampling frequency of the LIDS data is not set. '
                'Run lids_transform() before accessing this attribute.',
                UserWarning
            )
        return self.__freq

    @property
    def lids_func(self):
        r'''LIDS transformation function'''
        return self.__lids_func

    @property
    def lids_fit_func(self):
        r'''Fit function to LIDS oscillations'''
        return self.__fit_func

    @lids_fit_func.setter
    def lids_fit_func(self, func):
        self.__fit_func = func

    @property
    def lids_fit_initial_params(self):
        r'''Initial parameters of the fit function to LIDS oscillations'''
        return self.__fit_initial_params

    @property
    def lids_fit_results(self):
        r'''Results of the LIDS fit'''
        if self.__fit_results is None:
            warnings.warn(
                'The fit results is None. '
                'Run lids_fit() before accessing this attribute.',
                UserWarning
            )
        return self.__fit_results

    def filter(self, ts, duration_min='3H', duration_max='12H'):
        r'''Filter data according to their duration
        Before performing a LIDS analysis, it is necessary to drop sleep bouts
        that too short or too long.
        '''

        def duration(s):
            return s.index[-1] - s.index[0]

        td_min = pd.Timedelta(duration_min)
        td_max = pd.Timedelta(duration_max)

        from itertools import filterfalse
        filtered = []
        filtered[:] = filterfalse(
            lambda x: duration(x) < td_min or duration(x) > td_max,
            ts
        )
        return filtered

    def __smooth(self, lids, method, win_size):
        r'''Smooth LIDS data
        By default, smooth with a centered moving average using a `win_size`
        window'''

        # Smooth functions
        lids_smooth_funcs = ['mva', 'kernel', 'none']
        if method not in lids_smooth_funcs:
            raise ValueError(
                '`LIDS smooth function` must be "%s". You passed: "%s"' %
                ('" or "'.join(list(lids_smooth_funcs)), method)
            )

        if method == 'mva':
            return lids.rolling(win_size, center=True, min_periods=1).mean()
        elif method == 'kernel':
            smooth_lids = spm_smooth(lids.values, fwhm=win_size)
            return pd.Series(data=smooth_lids, index=lids.index)
        elif method == 'none':
            return lids

    def lids_transform(
            self, ts, method='mva', win_td='30min', resampling_freq=None
    ):
        r'''Apply LIDS transformation to activity data
        This transformation comprises:
        * resampling via summation (optional)
        * non-linear LIDS transformation
        * smoothing with a centered moving average
        Parameters
        ----------
        ts: pandas.Series
            Data identified as locomotor activity during sleep.
        method: str, optional
            Method to smooth the data.
            Available options are:
                'mva': moving average
                'kernel': gaussian kernel
                'none': no smoothing
            Default is 'mva'.
        win_td: str, optional
            Size of the moving average window.
            Default is '30min'.
        resampling_freq: str, optional
            Frequency of the resampling applied prior to LIDS transformation.
            Default is None.
        Returns
        -------
        smooth_lids: pandas.Series
        '''

        # Resample data to the required frequency
        if resampling_freq is not None:
            rs = ts.resample(resampling_freq).sum()
        else:
            rs = ts
        # Apply LIDS transformation x: 100/(x+1)
        lids = rs.apply(self.lids_func)

        # Store actual sampling frequency
        self.__freq = pd.Timedelta(lids.index.freq)

        # Series with a DateTimeIndex don't accept 'time-aware' centered window
        # Convert win_size (TimeDelta) into a number of time bins
        win_size = int(pd.Timedelta(win_td) / self.__freq)

        # Smooth LIDS-transformed data
        smooth_lids = self.__smooth(lids, method=method, win_size=win_size)

        return smooth_lids

    def lids_fit(
            self,
            lids,
            method='leastsq',
            scan_period=True,
            bounds=('30min', '180min'),
            step='5min',
            nan_policy='raise',
            verbose=False
    ):
        r'''Fit oscillations of the LIDS data
        The fit is performed with a fixed period ranging from 30 min to 180 min
        with a step of 5 min by default. The best-fit criterion is the maximal
        Munich Rhythmicity Index (MRI).
        Parameters
        ----------
        lids: pandas.Series
            Output data from LIDS transformation.
        method: str, optional
            Name of the fitting method to use [1]_.
            Default is 'leastsq'.
        scan_period: bool, optional
            If set to True, the period of the LIDS fit function is fixed and
            varied between the specified bounds.
            The selected period corresponds to the highest MRI value.
            Otherwise, the period is a free parameter of the fit.
            Default is True.
        bounds: 2-tuple of str, optional
            Lower and upper bounds for the periods to be tested.
            If scan_period is set to False, the bounds are ignored.
            Default is ('30min','180min').
        step: str, optional
            Time delta between the periods to be tested.
        nan_policy: str, optional
            Specifies action if the objective function returns NaN values.
            One of:
                'raise': a ValueError is raised
                'propagate': the values returned from userfcn are un-altered
                'omit': non-finite values are filtered
            Default is 'raise'.
        verbose: bool, optional
            If set to True, display fit informations
        References
        ----------
        .. [1] Non-Linear Least-Squares Minimization and Curve-Fitting for
               Python.
               https://lmfit.github.io/lmfit-py/index.html
        '''

        # # Define residual function to minimize
        # def residual(params, x, data):
        #     model = self.lids_fit_func(x, params)
        #     return (data-model)

        # Define the x range
        x = np.arange(lids.index.size)

        if scan_period:

            # Define bounds for the period
            period_start = pd.Timedelta(bounds[0]) / self.__freq
            period_end = pd.Timedelta(bounds[1]) / self.__freq
            period_range = period_end - period_start
            period_step = pd.Timedelta(step) / self.__freq

            test_periods = np.linspace(
                period_start,
                period_end,
                num=int(period_range / period_step) + 1
            )

            # Fit data for each test period
            mri = -np.inf
            fit_result_tmp = None
            initial_period = self.__fit_initial_params['period'].value
            for test_period in test_periods:
                # Fix test period
                self.__fit_initial_params['period'].value = test_period
                self.__fit_initial_params['period'].vary = False

                # Minimize residuals
                fit_result_tmp = minimize(
                    self.__fit_obj_func,
                    self.__fit_initial_params,
                    args=(x, lids.values, self.lids_fit_func),
                    nan_policy=nan_policy,
                    reduce_fcn=self.__fit_reduc_func
                )
                # Print fit parameters if verbose
                if verbose:
                    print(fit_report(fit_result_tmp))
                # Calculate the MR index
                mri_tmp = self.lids_mri(lids, fit_result_tmp.params)
                if verbose:
                    pearson_r = self.lids_pearson_r(
                        lids, fit_result_tmp.params)[0]
                    print('Pearson r: {}'.format(pearson_r))
                    print('MRI: {}'.format(mri_tmp))

                # If the newly calculated MRI is higher than the current MRI
                if mri_tmp > mri and (test_period != period_end):
                    # Store MRI
                    mri = mri_tmp
                    # Store fit parameters
                    fit_result = fit_result_tmp

            if verbose:
                print('Highest MRI: {}'.format(mri))
            # Set back original value
            self.__fit_initial_params['period'].value = initial_period
            self.__fit_initial_params['period'].vary = True
        else:

            # Minimize residuals
            fit_result = minimize(
                self.__fit_obj_func,
                self.__fit_initial_params,
                args=(x, lids.values, self.lids_fit_func),
                nan_policy=nan_policy,
                reduce_fcn=self.__fit_reduc_func
            )
            # Print fit parameters if verbose
            if verbose:
                print(fit_report(fit_result))
            if verbose:
                # Calculate the MR index
                pearson_r = self.lids_pearson_r(lids, fit_result.params)[0]
                mri = self.lids_mri(lids, fit_result.params)
                print('Pearson r: {}'.format(pearson_r))
                print('MRI: {}'.format(mri))

        self.__fit_results = fit_result
        # self.lids_fit_params = fit_result.params
        # self.lids_fit_period = fit_result.params['period'].value

    def lids_pearson_r(self, lids, params=None):
        r'''Pearson correlation factor
        Pearson correlation factor between LIDS data and its fit function
        Parameters
        ----------
        lids: pandas.Series
            Output data from LIDS transformation.
        params: lmfit.Parameters, optional
            Parameters for the fit function.
            If None, self.lids_fit_params is used instead.
            Default is None.
        Returns
        -------
        r: numpy.float64
            Pearson’s correlation coefficient
        p: numpy.float64
            2-tailed p-value
        '''

        x = np.arange(lids.index.size)
        if params is None:
            params = self.lids_fit_results.params
        return pearsonr(lids, self.lids_fit_func(x, params))

    def lids_mri(self, lids, params=None):
        r'''Munich Rhythmicity Index
        The Munich Rhythmicity Index (MRI) is defined as
        :math:`MRI = A \times r` with :math:`A`, the cosine fit amplitude and
        :math:`r`, the bivariate correlation coefficient (a.k.a. Pearson'r).
        Parameters
        ----------
        lids: pandas.Series
            Output data from LIDS transformation.
        params: lmfit.Parameters, optional
            Parameters for the fit function.
            If None, self.lids_fit_params is used instead.
            Default is None.
        Returns
        -------
        mri: numpy.float64
            Munich Rhythmicity Index
        '''
        if params is None:
            params = self.lids_fit_results.params

        # Pearson's r
        pearson_r = self.lids_pearson_r(lids, params)[0]

        # Oscillation range = [-A,+A] => 2*A
        oscillation_range = 2 * params['amp'].value

        # MRI
        mri = pearson_r * oscillation_range

        return mri

    def lids_period(self, freq='s'):
        r'''LIDS period
        Convert the period of the LIDS oscillations as estimated by the fit
        function to a TimeDelta.
        Parameters
        ----------
        s: str, optional
            Frequency to cast the output timedelta to.
            Default is 's'.
        Returns
        -------
        lids_period: numpy.timedelta64[freq]
        Note
        ----
        As there is no way to automatically derive the LIDS period from the fit
        parameters, the fitted period needs to be set via its own setter
        function.
        '''
        if self.freq is None:
            # TODO: evaluate if raise ValueError('') more appropriate
            return None
        elif self.lids_fit_results is None:
            # TODO: evaluate if raise ValueError('') more appropriate
            return None
        else:
            lids_period = self.lids_fit_results.params['period'] * self.freq
            return lids_period.astype('timedelta64[{}]'.format(freq))

    def lids_phases(self, lids, step=.1):
        r'''LIDS onset and offset phases in degrees
        Parameters
        ----------
        lids: pandas.Series
            Output data from LIDS transformation.
        step: float, optional
            Step between points at which the LIDS fit is evaluated.
            Default is '0.1'.
        Returns
        -------
        onset_phase, offset_phase: numpy.float64
        '''

        if self.lids_fit_results is None:
            # TODO: evaluate if raise ValueError('') more appropriate
            return None

        if self.lids_fit_results is None:
            # TODO: evaluate if raise ValueError('') more appropriate
            return None

        # Access fit parameters
        params = self.lids_fit_results.params

        # Fit support range
        x = np.arange(lids.index.size, step=step)

        # LIDS fit derivatives (1st and 2nd)
        df_dx = np.gradient(self.lids_fit_func(x, params), step)
        d2f_dx2 = np.gradient(df_dx, step)

        # Index of the 1st maxima (i.e 1st maximum of the LIDS oscillations)
        first_max_idx = np.argmax(_extrema_points(df_dx, d2f_dx2))
        # Convert the index into a phase using the fitted period
        onset_phase = (first_max_idx * step) / params['period'] * 360

        # Index of the last 'increasing' inflexion points in LIDS oscillations
        # before sleep offset
        last_inflex_idx = -1 * (
            # reverse order to find last
                np.argmax(_inflexion_points(df_dx, d2f_dx2)[::-1]) +
                1  # to account for index shifting during reverse (-1: 0th elem)
        )
        # Convert the index into a phase using the fitted period
        offset_phase = np.abs(last_inflex_idx * step / params['period'] * 360)

        return onset_phase, offset_phase

    def lids_convert_to_internal_time(self, lids, t_norm='90min'):
        r'''Convert LIDS data index to internal time.
        XXX
        Parameters
        ----------
        lids: pandas.Series
            Output data from LIDS transformation.
        t_norm: str, optional
            Default period used to normalize the fitted LIDS period.
            Default is '90min'.
        Returns
        -------
        ts: pandas.Series
            LIDS data with internal time since sleep onset as index.
        '''

        # External timeline of the current LIDS data since sleep onset
        t_ext = pd.timedelta_range(
            start='0 day',
            periods=lids.index.size,
            freq=self.freq
        )

        # Scaling factor, relative to the LIDS period, normalized to t_norm
        scaling_factor = pd.Timedelta(t_norm) / self.lids_period()

        # Internal timeline (aka: external timeline, rescaled to LIDS period)
        # t_int = scaling_factor*t_ext
        t_int = pd.TimedeltaIndex(scaling_factor * t_ext.values, freq='infer')

        # Construct a new Series with internal timeline as index
        lids_rescaled = pd.Series(lids.values, index=t_int)

        # Resample LIDS data to restore the original bin width of its Index
        # Infer missing data via interpolation
        lids_resampled = lids_rescaled.resample(
            self.freq
            # label='right',
            # closed='right'
        ).mean()

        return lids_resampled.interpolate(method='linear')

    def lids_summary(self, lids, verbose=False):
        r'''Calculate summary statistics for LIDS
        Fit all LIDS-transformed bouts and calculate the mean period, the mean
        mri, the mean number of LIDS cycles and the dampening factor of the
        mean LIDS profile.
        Parameters
        ----------
        lids: list of pandas.Series
            Output data from LIDS transformation.
        verbose: bool, optional
            If set to True, print summary statistics.
            Default is False.
        Returns
        -------
        summary: dict
            Dictionary with the summary statistics.
        '''

        ilids = []  # LIDS profiles
        periods = []  # List of LIDS periods
        mris = []  # MRI indices
        ncycles = []  # Number of LIDS cycles/sleep bout

        for idx, s in enumerate(lids):
            # Fit LIDS data
            self.lids_fit(s, verbose=False)

            # Verify LIDS period
            period = self.lids_period(freq='s')

            # Calculate MRI
            mri = self.lids_mri(s)

            # Calculate the number of LIDS cycle (as sleep bout length/period):
            ncycle = s.index.values.ptp() / np.timedelta64(1, 's')
            ncycle /= period.astype(float)

            if verbose:
                print('-' * 20)
                print('Sleep bout nr {}'.format(idx))
                print('- Period: {!s}'.format(period))
                print('- MRI: {}'.format(mri))
                print('- Number of cycles: {}'.format(ncycle))

            # Rescale LIDS timeline to LIDS period
            rescaled_lids = self.lids_convert_to_internal_time(s)

            periods.append(period)
            mris.append(mri)
            ncycles.append(ncycle)
            ilids.append(rescaled_lids)

        # Create the mean LIDS profile
        lids_profile = reduce(
            (lambda x, y: x.add(y, fill_value=0)),
            ilids
        ) / len(ilids)

        # Fit mean LIDS profile with a pol0
        fit_params = np.polyfit(
            x=range(len(lids_profile.index)),
            y=lids_profile.values,
            deg=1
        )

        # LIDS summary
        summary = {}
        summary['Mean number of LIDS cycles'] = np.mean(ncycles)
        summary['Mean LIDS period (s)'] = np.mean(periods).astype(float)
        summary['Mean MRI'] = np.mean(mris)
        summary[
            'LIDS dampening factor (counts/{})'.format(self.freq)
        ] = fit_params[0]

        return summary



from glob import glob
import numpy as np
import os
import warnings

from hypnospy.data import RawProcessing
from hypnospy import Wearable
from hypnospy import Diary


class Experiment(object):

    def __init__(self):
        self.datapath = None
        self.wearables = {}

    def add_wearable(self, w):
        self.wearables[w.get_pid()] = w

    def remove_wearable(self, w):
        if w in self.wearables:
            del self.wearables[w]

    def configure_experiment(self, datapath: str, device_location: str, cols_for_activity, col_for_mets: str = None,
                             is_emno: bool = False, is_act_count: bool = False, col_for_datetime: str = "time",
                             start_of_week: int = -1, strftime: str = None, col_for_pid: str = None, pid: int = -1,
                             additional_data: object = None, col_for_hr: str = None):

        # TODO: Missing a check to see if datapath exists.
        if os.path.isdir(datapath):
            if not datapath.endswith("*"):
                datapath = os.path.join(datapath, "*")
        else:
            if '*' not in datapath:
                datapath = datapath + "*"

        for file in glob(datapath):
            pp = RawProcessing(file,
                               device_location=device_location,
                               # activitiy information
                               cols_for_activity=cols_for_activity,
                               is_act_count=is_act_count,
                               is_emno=is_emno,
                               # Datatime information
                               col_for_datetime=col_for_datetime,
                               col_for_mets=col_for_mets,
                               start_of_week=start_of_week,
                               strftime=strftime,
                               # Participant information
                               col_for_pid=col_for_pid,
                               pid=pid,
                               additional_data=additional_data,
                               # HR information
                               col_for_hr=col_for_hr, )
            w = Wearable(pp)
            self.wearables[w.get_pid()] = w

    def add_diary(self, diary: Diary):
        for pid in diary.data["pid"].unique():
            di = diary.data[diary.data["pid"] == pid]
            w = self.get_wearable(pid)
            if w:
                w.add_diary(Diary().from_dataframe(di))

    def get_wearable(self, pid: str):
        if pid in self.wearables.keys():
            return self.wearables[pid]
        else:
            warnings.warn("Unknown PID %s." % pid)
            return None
            # raise KeyError("Unknown PID %s." % pid)

    def size(self):
        return len(self.wearables)

    def get_all_wearables(self):
        return list(self.wearables.values())

    def set_freq_in_secs(self, freq: int):
        for wearable in self.get_all_wearables():
            wearable.set_frequency_in_secs(freq)

    def change_start_hour_for_experiment_day(self, hour):
        for wearable in self.get_all_wearables():
            wearable.change_start_hour_for_experiment_day(hour)

    def fill_no_activity(self, value):
        for wearable in self.get_all_wearables():
            wearable.fill_no_activity(value)

    def overall_stats(self):
        days_acc = []
        epochs_acc = []
        for wearable in self.get_all_wearables():
            days_acc.append(len(wearable.data[wearable.get_experiment_day_col()].unique()))
            epochs_acc.append(wearable.data.shape[0])

        days_acc = np.array(days_acc)
        epochs_acc = np.array(epochs_acc)
        print("Total number of wearables: %d" % (self.size()))
        if epochs_acc.size > 0:
            print("Avg. number of days: %.2f (+-%.3f). Max: %d, Min: %d." % (days_acc.mean(), days_acc.std(), days_acc.max(), days_acc.min()))
            print("Avg. number of epochs: %.2f (+-%.3f). Max: %d, Min: %d."% (epochs_acc.mean(), epochs_acc.std(), epochs_acc.max(), epochs_acc.min()))
        else:
            print("Experiment has no valid wearable left.")


from hypnospy import Wearable
from hypnospy import Experiment
from hypnospy import misc

import pandas as pd


class PhysicalActivity(object):

    def __init__(self, input: {Experiment, Wearable}, cutoffs=None, names=None):
        """

        :param input: Either an experiment or wearable object.
        :param cutoffs: List of cut-offs
        :param names: List of physical activity names associated with cut-offs.
        Note: it is expected that len(names) will be len(cutoffs) + 1
        """

        if type(input) is Wearable:
            self.wearables = [input]
        elif type(input) is Experiment:
            self.wearables = input.get_all_wearables()

        self.set_cutoffs(cutoffs, names)

    def set_cutoffs(self, cutoffs, names):
        """
        This method is used to define physical activity cut-offs and their respective names.

        The cut-off values here are tightly related to the wearable device used.
        We suggest the user to read the latest research on it.
        Vincent van Hees' GGIR has a summarized documentation on this topic: see https://cran.r-project.org/web/packages/GGIR/vignettes/GGIR.html#published-cut-points-and-how-to-use-them


        :param cutoffs: List of values
        :param names:  List of names. Expected to have one element more than ``cutoffs``.
        :return: None
        """
        if cutoffs is None and names is None:
            return

        # N cut-offs defined N+1 classes
        assert len(cutoffs) == len(names) - 1

        # Check if cutoffs are in increasing order
        assert sorted(cutoffs) == cutoffs

        self.cutoffs = cutoffs
        self.names = names

    def generate_pa_columns(self, based_on="activity"):
        """

        :param based_on: Base column used to calculate the physical activity.
        :return: None
        """

        if self.cutoffs is None or self.names is None:
            raise AttributeError("Please use `set_cutoffs` before using this method.")

        for wearable in self.wearables:
            # first cut-off
            wearable.data[self.names[0]] = wearable.data[based_on] <= self.cutoffs[0]
            # Other cut-ffs
            for i in range(1, len(self.names)):
                wearable.data[self.names[i]] = wearable.data[based_on] > self.cutoffs[i - 1]

            wearable.pa_cutoffs = self.cutoffs
            wearable.pa_names = self.names

    def get_bouts(self, pa_col, length_in_minutes, decomposite_bouts=False):
        """

        :param pa_col:              The name of the physical activity column in the dataframe.
        :param length_in_minutes:   The minimal length of the activity in minutes
        :param decomposite_bouts:   If True, we are going to count the number of subsequencies of ``length_in_minutes``.
                                    For example, if ``length_in_minutes`` is 10 and we identified a continuous activity of 30 minutes,
                                    if ``decomposite_bouts`` is True, we will count 3 bouts for this activity, otherwise only 1.
        :return:                    A dictionary with <pid> keys and Series <day, bouts> as values.
        """

        if pa_col not in self.names:
            raise ValueError("Unknown physical activity column %s. Please use ``set_cutoffs``.")

        returning_dict = {}
        for wearable in self.wearables:
            pid = wearable.get_pid()
            epochs_per_minute = wearable.get_epochs_in_min()
            returning_dict[pid] = wearable.data.groupby(wearable.experiment_day_col).apply(
                lambda x: self._get_bout(x, epochs_per_minute, pacol=pa_col, mins=length_in_minutes,
                                         decomposite_bouts=decomposite_bouts)
            )
        return returning_dict

    @staticmethod
    def _get_bout(dd, epochs_per_minute, pacol, mins=10, decomposite_bouts=True):

        dd["pa_len"], dd["pa_grp"] = misc.get_consecutive_serie(dd, pacol)
        bouts = dd[(dd[pacol] == True) & (dd["pa_len"] >= mins * epochs_per_minute)]

        if decomposite_bouts:
            # Should we break a 20 min bout in 2 of 10 mins?
            bouts["num_sub_bouts"] = bouts["pa_len"] // (mins * epochs_per_minute)
            return bouts.groupby("pa_grp").first()["num_sub_bouts"].sum()
        else:
            return len(bouts["pa_grp"].unique())

    def mask_sleep_period(self, mask):
        # TODO: missing implementation and finding a better name for this method.
        df["X"] = df["activity"].where(~df["sleep_period"].astype(np.bool), 0)

    def get_binned_pa_representation(self):
        """

        :return:
        """
        rows = []
        for wearable in self.wearables:
            act_hour = \
                wearable.data.groupby(
                    [wearable.get_experiment_day_col(), wearable.data[wearable.get_time_col()].dt.hour])[
                    wearable.get_activity_col()]

            pid = wearable.get_pid()
            # lpa = act_hour.apply(lambda x: (x <= self.mvpa_value).sum())
            lpa = act_hour.apply(lambda x: (x <= self.cutoffs[0]).sum())
            lpa.name = "LPA"
            # mvpa = act_hour.apply(lambda x: x.between(self.mvpa_value, self.vpa_value).sum())
            mvpa = act_hour.apply(lambda x: x.between(self.cutoffs[0], self.cutoffs[1]).sum())
            mvpa.name = "MVPA"
            # vpa = act_hour.apply(lambda x: (x >= self.vpa_value).sum())
            vpa = act_hour.apply(lambda x: (x >= self.cutoffs[1]).sum())
            vpa.name = "VPA"

            concatenated = pd.concat([lpa, mvpa, vpa], axis=1)
            concatenated["pid"] = pid
            rows.append(concatenated)

        return pd.concat(rows)

    def get_stats_pa_representation(self):
        """

        :return: a dataframe with statistical measures by day and hour of the experiment
        """
        rows = []
        for wearable in self.wearables:
            act_hour = \
                wearable.data.groupby(
                    [wearable.get_experiment_day_col(), wearable.data[wearable.get_time_col()].dt.hour])[
                    wearable.get_activity_col()]

            rows.append(pd.DataFrame({"pid": wearable.get_pid(), "mean": act_hour.mean(),
                                      "median": act_hour.median(),
                                      "std": act_hour.std(),
                                      "min": act_hour.min(),
                                      "max": act_hour.max(),
                                      "skewness": act_hour.skew(),
                                      "kurtosis": act_hour.apply(pd.Series.kurt),
                                      "nunique": act_hour.nunique(),
                                      # Could give an indication of variability in the group
                                      }))
        return pd.concat(rows)

    def get_raw_pa(self):
        #  TODO: actually, one interesting way to implement it is having a concept of experiment_day that starts from 1
        #   rather than the current experiment_day that start from the day of the week

        # rows = []
        # for wearable in self.wearables:
        #     act_hour = \
        #         wearable.data.groupby(
        #             [wearable.get_experiment_day_col(), wearable.data[wearable.get_time_col()].dt.hour])[
        #             wearable.get_activity_col()]
        #     #act_hour.apply(lambda x: x.values.ravel())
        #     #act_hour["pid"] = wearable.get_pid()
        #     #rows.append(act_hour)
        #     return act_hour
        # #return pd.concat(rows)
        pass



from hypnospy import Wearable
from hypnospy import Experiment
from hypnospy import misc

from sklearn import metrics
import pandas as pd


class SleepMetrics(object):

    def __init__(self, input: {Experiment, Wearable}):

        if type(input) is Wearable:
            self.wearables = [input]
        elif type(input) is Experiment:
            self.wearables = input.get_all_wearables()

    @staticmethod
    def calculate_sleep_efficiency(df_in, wake_col, wake_delay_epochs=0):
        if wake_delay_epochs == 0:
            return 100 * (1. - df_in[wake_col].sum() / df_in.shape[0])

        else:
            # Avoid modifying the original values in the wake col
            df = df_in.copy()
            df["consecutive_state"], _ = misc.get_consecutive_serie(df, wake_col)
            # Change wake from 1 to 0 if group has less than ``wake_delay_epochs`` entries
            df.loc[(df[wake_col] == 1) & (df["consecutive_state"] <= wake_delay_epochs), wake_col] = 0
            sleep_quality = 100 * (1. - df[wake_col].sum() / df.shape[0])
            # delete aux cols
            return sleep_quality

    @staticmethod
    def calculate_awakening(df_in, wake_col, wake_delay=0, normalize_per_hour=False, epochs_in_hour=0):
        df = df_in.copy()

        df["consecutive_state"], df["gids"] = misc.get_consecutive_serie(df, wake_col)

        grps = df[(df[wake_col] == 1) & (df["consecutive_state"] > wake_delay)].groupby("gids")
        del df["consecutive_state"]
        del df["gids"]

        if normalize_per_hour:
            total_hours_slept = df.shape[0] / epochs_in_hour
            return len(grps) / total_hours_slept
        else:
            return len(grps)

    @staticmethod
    def calculate_arousals(df, wake_col=None, normalize_per_hour=False, epochs_in_hour=0):

        arousals = ((df[wake_col] == 1) & (df[wake_col] != df[wake_col].shift(1).fillna(0))).sum()

        if normalize_per_hour:
            total_hours_slept = df.shape[0] / epochs_in_hour
            return arousals / total_hours_slept
        else:
            return arousals

    @staticmethod
    def calculate_sri(prev_block, current_block, wake_col):

        if prev_block.shape[0] != current_block.shape[0]:
            raise ValueError("Unable to calculate SRI.")

        same = (prev_block[wake_col].values == current_block[wake_col].values).sum()
        sri = (same / prev_block.shape[0]) * 100.

        return sri

    def compare_sleep_metrics(self, ground_truth, wake_sleep_alg, sleep_metrics=["sleepEfficiency", "awakening", "SRI",
                                                                                 "arousal", "totalTimeInBed",
                                                                                 "totalSleepTime", "totalWakeTime"],
                              how="relative_difference"):

        results = []
        for sleep_metric in sleep_metrics:

            gt = self.get_sleep_quality(wake_col=ground_truth, metric=sleep_metric)
            other = self.get_sleep_quality(wake_col=wake_sleep_alg, metric=sleep_metric)

            gtdf = pd.DataFrame(gt)
            otherdf = pd.DataFrame(other)
            merged = pd.merge(gtdf, otherdf, on=["pid", "expday", "metric"], suffixes=["_gt", "_other"])

            if how == "relative_difference":
                merged["value"] = merged[["value_gt", "value_other"]].apply(
                    lambda x: ((x["value_gt"] - x["value_other"]) / x["value_other"]) * 100. if x[
                                                                                                    "value_other"] is not None and
                                                                                                x[
                                                                                                    "value_other"] > 0 else 0,
                    axis=1)
                merged["metric"] = "delta_" + merged["metric"]
                merged["alg1"] = ground_truth
                merged["alg2"] = wake_sleep_alg
                del merged["value_gt"]
                del merged["value_other"]

                results.append(merged)

            elif how == "pearson":
                value = merged[["value_gt", "value_other"]].corr("pearson")["value_gt"]["value_other"]

                s = pd.Series({"value": value, "metric": "pearson_" + sleep_metric,
                               "alg1": ground_truth, "alg2": wake_sleep_alg})
                results.append(s)

        if how == "relative_difference":
            concated = pd.concat(results)
            return concated.to_dict('records')

        else:
            return pd.concat(results, axis=1).T.to_dict('records')

    def compare(self, ground_truth, wake_sleep_alg, sleep_period_col=None):

        results = {}

        # Usual evaluation metrics such as Accuracy, Precision, F1....
        for wearable in self.wearables:
            df = wearable.data
            result = {"accuracy": [], "precision": [], "recall": [], "f1_score": [], "roc_auc": [], "cohens_kappa": []}

            for day, block in df.groupby(wearable.experiment_day_col):

                # Filter where we should calculate the sleep metric
                if sleep_period_col is not None:
                    block = block[block[sleep_period_col] == True]

                gt, pred = block[ground_truth], block[wake_sleep_alg]

                result["accuracy"].append(metrics.accuracy_score(gt, pred))
                result["precision"].append(metrics.precision_score(gt, pred))
                result["recall"].append(metrics.recall_score(gt, pred))
                result["f1_score"].append(metrics.f1_score(gt, pred))
                result["roc_auc"].append(metrics.roc_auc_score(gt, pred))
                result["cohens_kappa"].append(metrics.cohen_kappa_score(gt, pred))

            results[wearable.get_pid()] = result

        return results

    def get_sleep_quality(self, wake_col, sleep_period_col=None, metric="sleepEfficiency", wake_delay_in_minutes=10):
        """
            This function implements different notions of sleep quality.
            For far strategy can be:
            - sleepEfficiency (0-100): the percentage of time slept (wake=0) in the dataframe
            - awakening (> 0): counts the number of awakenings in the period. An awakening is a sequence of wake=1 periods larger than th_awakening (default = 10)
            - awakeningIndex (> 0)
            - arousal (> 0):
            - arousalIndex (>0):
            - totalTimeInBed (in hours)
            - totalSleepTime (in hours)
            - totalWakeTime (in hours)
            - SRI (Sleep Regularity Index, in percentage %)
        """

        results = []
        for wearable in self.wearables:
            df = wearable.data
            wake_delay_in_epochs = wearable.get_epochs_in_min() * wake_delay_in_minutes
            first_day = True
            prev_block = None

            for day, block in df.groupby(wearable.experiment_day_col):

                row = {"pid": wearable.get_pid(), "expday": day}

                # filter where we should calculate the sleep metric
                if sleep_period_col is not None:
                    block = block[block[sleep_period_col] == True]

                if metric.lower() in ["sleepefficiency", "se"]:
                    row["metric"] = "sleepefficiency"
                    row["value"] = SleepMetrics.calculate_sleep_efficiency(block, wake_col,
                                                                           wake_delay_in_epochs)

                elif metric.lower() in ["awakening", "awakenings"]:
                    row["metric"] = "awakening"
                    row["value"] = SleepMetrics.calculate_awakening(block, wake_col,
                                                                    wake_delay_in_epochs,
                                                                    normalize_per_hour=False)

                elif metric == "awakeningIndex":
                    row["metric"] = "awakeningIndex"
                    row["value"] = SleepMetrics.calculate_awakening(block, wake_col, wake_delay_in_epochs,
                                                                    normalize_per_hour=True,
                                                                    epochs_in_hour=wearable.get_epochs_in_hour())

                elif metric == "arousal":
                    row["metric"] = "arousal"
                    row["value"] = SleepMetrics.calculate_arousals(block, wake_col, normalize_per_hour=False)

                elif metric == "arousalIndex":
                    row["metric"] = "arousalIndex"
                    row["value"] = SleepMetrics.calculate_arousals(block, wake_col, normalize_per_hour=True,
                                                                   epochs_in_hour=wearable.get_epochs_in_hour())
                elif metric == "totalTimeInBed":
                    row["metric"] = "totalTimeInBed"
                    row["value"] = block.shape[0] / wearable.get_epochs_in_hour()

                elif metric == "totalSleepTime":
                    row["metric"] = "totalSleepTime"
                    row["value"] = ((block[wake_col] == 0).sum()) / wearable.get_epochs_in_hour()

                elif metric == "totalWakeTime":
                    row["metric"] = "totalWakeTime"
                    row["value"] = (block[wake_col].sum()) / wearable.get_epochs_in_hour()

                elif metric.lower() in ["sri", "sleep_regularity_index", "sleepregularityindex"]:
                    if first_day:
                        first_day = False
                        prev_block = block
                        continue

                    row["metric"] = "sri"
                    try:
                        sri = SleepMetrics.calculate_sri(prev_block, block, wake_col)
                    except ValueError:
                        print("Unable to calculate SRI for day %d (PID = %s)." % (day, wearable.get_pid()))
                        sri = None
                    row["value"] = sri
                    prev_block = block

                else:
                    raise ValueError("Metric %s is unknown." % metric)

                results.append(row)

        return pd.DataFrame(results)

    def __evaluate_sleep_boundaries_pair(self, ground_truth, other):

        df_acc = []
        expid = 0

        for w in self.wearables:

            if w.data.empty:
                print("Data for PID %s is empty!" % w.get_pid())
                continue

            if ground_truth not in w.data:
                print("Column %s not in dataset for PID %s." % (ground_truth, w.get_pid()))
                continue

            if other not in w.data:
                print("Column %s not in dataset for PID %s." % (other, w.get_pid()))
                continue

            sleep = {}
            sleep[ground_truth] = w.data[ground_truth].astype(int)
            sleep[other] = w.data[other].astype(int)

            if sleep[ground_truth].shape[0] == 0:
                continue

            mse = metrics.mean_squared_error(sleep[ground_truth], sleep[other])
            cohen = metrics.cohen_kappa_score(sleep[ground_truth], sleep[other])

            tst_gt = w.get_total_sleep_time_per_day(sleep_col=ground_truth)
            tst_gt.rename(columns={ground_truth: "tst_" + ground_truth}, inplace=True)
            tst_other = w.get_total_sleep_time_per_day(sleep_col=other)
            tst_other.rename(columns={other: "tst_" + other}, inplace=True)

            onset_gt = w.get_onset_sleep_time_per_day(sleep_col=ground_truth)
            onset_gt.name = "onset_" + ground_truth
            onset_other = w.get_onset_sleep_time_per_day(sleep_col=other)
            onset_other.name = "onset_" + other

            offset_gt = w.get_offset_sleep_time_per_day(sleep_col=ground_truth)
            offset_gt.name = "offset_" + ground_truth
            offset_other = w.get_offset_sleep_time_per_day(sleep_col=other)
            offset_other.name = "offset_" + other

            df_res = pd.concat((onset_gt, onset_other, offset_gt, offset_other, tst_gt, tst_other), axis=1)
            df_res["pid"] = w.get_pid()
            df_res["mse_" + ground_truth + "&" + other] = mse
            df_res["cohens_" + ground_truth + "&" + other] = cohen
            df_res = df_res.reset_index()
            expid += 1

            df_acc.append(df_res)

        df_acc = pd.concat(df_acc)

        # Drop invalid rows:
        for col in ["onset_", "offset_", "tst_"]:
            df_acc = df_acc[~df_acc[col + ground_truth].isnull()]

        return df_acc

    def evaluate_sleep_boundaries(self, ground_truth, others):
        dfs = []
        for other in others:
            dfs.append(self.__evaluate_sleep_boundaries_pair(ground_truth, other))

        if len(dfs) == 0:
            return None

        result = dfs[0]
        for other in dfs[1:]:
            common_keys = list(set(result.keys()).intersection(other.keys()))
            result = result.merge(other, on=common_keys)

        return result



from unittest import TestCase

from hypnospy.data import RawProcessing
from hypnospy import Wearable
from hypnospy.analysis import TimeSeriesProcessing


class TestWearable(TestCase):

    def setUp(self):
        # Loads some file
        pp1 = RawProcessing()
        pp1.load_file("../data/examples_mesa/mesa-sample-day5-invalid5hours.csv",
                      # activitiy information
                      cols_for_activity=["activity"],
                      is_act_count=True,
                      # Datatime information
                      col_for_datatime="linetime",
                      device_location="dw",
                      start_of_week="dayofweek",
                      # Participant information
                      col_for_pid="mesaid")
        self.w_5day_invalid5hours = Wearable(pp1)

    def test_get_activity_col(self):
        col_name = self.w_5day_invalid5hours.get_activity_col()
        self.assertEqual(col_name, "hyp_act_x")
        self.assertIsInstance(col_name, str)

    def test_get_pid(self):
        pid = self.w_5day_invalid5hours.get_pid()
        self.assertEqual(pid, "1")
        self.assertIsInstance(pid, str)

    def test_get_experiment_day_col(self):
        exp_day_col = self.w_5day_invalid5hours.get_experiment_day_col()
        self.assertEqual(exp_day_col, "hyp_exp_day")
        self.assertIsInstance(exp_day_col, str)

    def test_get_time_col(self):
        time_col = self.w_5day_invalid5hours.get_time_col()
        self.assertEqual(time_col, "hyp_time")
        self.assertIsInstance(time_col, str)

    def test_get_frequency_in_secs(self):
        freq = self.w_5day_invalid5hours.get_frequency_in_secs()
        self.assertEqual(freq, 30)
        self.assertIsInstance(freq, int)

    def test_get_epochs_in_min(self):
        nepochs = self.w_5day_invalid5hours.get_epochs_in_min()
        self.assertEqual(nepochs, 2.0)
        self.assertIsInstance(nepochs, float)

    def test_get_epochs_in_hour(self):
        nepochs = self.w_5day_invalid5hours.get_epochs_in_hour()
        self.assertEqual(nepochs, 120.0)
        self.assertIsInstance(nepochs, float)

    def test_change_start_hour_for_experiment_day(self):
        self.w_5day_invalid5hours.change_start_hour_for_experiment_day(0)

        # We are expecting to have only one experiment day and this will be day 5
        self.assertEqual(self.w_5day_invalid5hours.data["hyp_exp_day"].unique()[0], 5)

        # We now start our day at hour 18
        self.w_5day_invalid5hours.change_start_hour_for_experiment_day(18)
        # print(tsp.wearable.data[2155:2165])
        # Check if transition from artificial day 4 to day 5 is done correctly
        self.assertEqual(self.w_5day_invalid5hours.data.iloc[2159]["hyp_exp_day"], 4)
        self.assertEqual(self.w_5day_invalid5hours.data.iloc[2160]["hyp_exp_day"], 5)

        # Randomly change the start hour and return it back to 18
        self.w_5day_invalid5hours.change_start_hour_for_experiment_day(18)
        self.w_5day_invalid5hours.change_start_hour_for_experiment_day(0)
        self.w_5day_invalid5hours.change_start_hour_for_experiment_day(15)
        self.w_5day_invalid5hours.change_start_hour_for_experiment_day(18)
        self.assertEqual(self.w_5day_invalid5hours.data.iloc[2159]["hyp_exp_day"], 4)
        self.assertEqual(self.w_5day_invalid5hours.data.iloc[2160]["hyp_exp_day"], 5)

    def test_has_nan_activity(self):
        self.assertTrue(self.w_5day_invalid5hours.has_no_activity())
        self.w_5day_invalid5hours.fill_no_activity(1)
        self.assertFalse(self.w_5day_invalid5hours.has_no_activity())

    def test_valid_invalid_days(self):
        # Should not return anything yet, as we never marked any row as invalid
        invalid_days = self.w_5day_invalid5hours.get_invalid_days()
        self.assertSetEqual(invalid_days, set())

        valid_days = self.w_5day_invalid5hours.get_valid_days()
        self.assertSetEqual(valid_days, set([5]))

        #  We now force some an invalid day
        tsp = TimeSeriesProcessing(self.w_5day_invalid5hours)
        tsp.detect_non_wear(strategy="choi2011")
        tsp.check_valid_days(min_activity_threshold=0, max_non_wear_minutes_per_day=60)

        invalid_days = self.w_5day_invalid5hours.get_invalid_days()
        self.assertSetEqual(invalid_days, set({5}))



import hypnospy
import pandas as pd
from hypnospy import Diary
import h5py


class Wearable(object):

    def __init__(self, input):
        """

        :param input: Either a path to a PreProcessing file saved with ``export_hyp`` or a PreProcessing object
        """
        self.data = None
        self.freq_in_secs = None
        self.hour_start_experiment = None
        self.experiment_day_col = "hyp_exp_day"

        # Other fields
        self.mets_col = None
        self.device_location = None
        self.additional_data = None
        # Participant Info
        self.pid = None
        # Activity Info
        self.activitycols = None
        self.mets_col = None
        self.is_act_count = None
        self.is_emno = None
        # Time Info
        self.time_col = None
        # HR Info
        self.hr_col = None
        # Diary
        self.diary = None
        self.diary_onset = "hyp_diary_onset"
        self.diary_offset = "hyp_diary_offset"
        self.diary_sleep = "hyp_diary_sleep"

        if isinstance(input, str):
            # Reads a hypnosys file from disk
            self.__read_hypnospy(input)

        elif isinstance(input, hypnospy.data.preprocessing.RawProcessing):
            self.__read_preprocessing_obj(input)
            # print("Loaded wearable with pid %s" % (self.get_pid()))

        # Creates the experiment day and set it the initial hour to be midnight
        self.change_start_hour_for_experiment_day(0)

    def __read_preprocessing_obj(self, input):
        "input is a wearable object. We can copy its fields"
        self.data = input.data
        self.device_location = input.device_location
        self.additional_data = input.additional_data
        # Participant Info
        self.pid = str(input.pid)
        # Activity Info
        self.activitycols = input.internal_activity_cols
        self.mets_col = input.internal_mets_col
        self.is_act_count = input.is_act_count
        self.is_emno = input.is_emno
        # Time Info
        self.time_col = input.internal_time_col
        # HR Info
        self.hr_col = input.internal_hr_col

    def __read_hypnospy(self, filename):

        self.data = pd.read_hdf(filename, 'data')
        l = pd.read_hdf(filename, 'other')
        self.pid, self.time_col, self.activitycols, self.internal_mets_col, self.is_act_count, self.is_emno, \
        self.device_location, self.additional_data = l

        self.pid = str(self.pid)

        # hf = h5py.File(filename, 'r')
        # self.data = hf.get('data')
        # self.device_location = hf.get('location')
        # self.additional_data = hf.get('additional_data')

    def get_pid(self):
        return self.pid

    def get_experiment_day_col(self):
        return self.experiment_day_col

    def get_mets_col(self):
        return self.mets_col

    def get_time_col(self):
        return self.time_col

    def set_frequency_in_secs(self, freq):
        self.freq_in_secs = freq

    def get_frequency_in_secs(self):
        if self.freq_in_secs:
            return self.freq_in_secs

        freq_str = pd.infer_freq(self.data[self.time_col])
        if freq_str is None:
            raise ValueError("Could not infer the frequency for pid %s." % self.get_pid())
        # pd.to_timedelta requires we have a number
        if not freq_str[0].isdigit():
            freq_str = "1" + freq_str
        return int(pd.to_timedelta(freq_str).total_seconds())

    def get_epochs_in_min(self):
        # TODO: should we force it to be integer?
        return 60 / self.get_frequency_in_secs()

    def get_epochs_in_hour(self):
        # TODO: should we force it to be integer?
        return 60 * self.get_epochs_in_min()

    def fill_no_activity(self, value):
        # TODO: write the use case for triaxial devices.
        self.data[self.get_activity_col()].fillna(value, inplace=True)

    def has_no_activity(self):
        return self.data[self.get_activity_col()].isnull().any()

    def change_start_hour_for_experiment_day(self, hour_start_experiment):
        """
        Allows the experiment to start in another time than 00:00.

        :param hour_start_experiment: 0: midnight, 1: 01:00AM ...
        """
        self.hour_start_experiment = hour_start_experiment
        day_zero = self.data.iloc[0][self.time_col].toordinal()
        new_exp_day = (self.data[self.time_col] - pd.DateOffset(hours=hour_start_experiment)).apply(
            lambda x: x.toordinal() - day_zero)

        self.data[self.experiment_day_col] = new_exp_day

    def get_activity_col(self):
        return self.activitycols[0]

    def get_hr_col(self):
        return self.hr_col

    def add_diary(self, d: Diary):
        d.data = d.data[d.data["pid"] == self.get_pid()]
        self.diary = d
        self.data[self.diary_onset] = False
        self.data[self.diary_offset] = False
        self.data.loc[self.data[self.time_col].isin(self.diary.data["sleep_onset"]), self.diary_onset] = True
        self.data.loc[self.data[self.time_col].isin(self.diary.data["sleep_offset"]), self.diary_offset] = True

        self.data[self.diary_sleep] = False
        for _, row in self.diary.data.iterrows():
            if not pd.isna(row["sleep_onset"]) and not pd.isna(row["sleep_offset"]):
                self.data.loc[(self.data[self.time_col] >= row["sleep_onset"]) & (
                        self.data[self.time_col] <= row["sleep_offset"]), self.diary_sleep] = True

    def get_total_sleep_time_per_day(self, sleep_col: str = None, based_on_diary: bool = False):
        """

        :param sleep_col:
        :param based_on_diary:
        :return: A Series indexed by experiment_day with total of minutes slept per day
        """

        if not based_on_diary and sleep_col is None:
            raise ValueError("Unable to calculate total sleep time."
                             " You have to specify a sleep column or set ``based_on_diary`` to True "
                             "(assuming you previously added a diary.")
        if based_on_diary:
            if self.diary is None:
                raise ValueError(
                    "Diary not found for PID %s. Add a diary with ``Wearable.add_diary``." % (self.get_pid()))
            return self.data.groupby(self.experiment_day_col)[[self.diary_sleep]].apply(
                lambda x: x.sum() / self.get_epochs_in_min())

        else:
            if sleep_col not in self.data.keys():
                raise ValueError(
                    "Could not find sleep_col named %s for PID %s. Aborting." % (self.get_pid(), sleep_col))
            return self.data.groupby(self.experiment_day_col)[[sleep_col]].apply(
                lambda x: x.sum() / self.get_epochs_in_min())

    def get_onset_sleep_time_per_day(self, sleep_col: str = None, based_on_diary: bool = False):
        """

        :param sleep_col:
        :param based_on_diary:
        :return: A Series indexed by experiment_day with total of minutes slept per day
        """

        if not based_on_diary and sleep_col is None:
            raise ValueError("Unable to calculate total sleep time."
                             " You have to specify a sleep column or set ``based_on_diary`` to True "
                             "(assuming you previously added a diary.")
        if based_on_diary:
            if self.diary is None:
                raise ValueError("Diary not found. Add a diary with ``add_diary``.")
            event = self.data[self.data[self.diary_onset] == True]
        else:
            if sleep_col not in self.data.keys():
                raise ValueError("Could not find sleep_col (%s). Aborting." % sleep_col)
            event = self.data[self.data[sleep_col] == True]

        return event.groupby(self.experiment_day_col)[self.time_col].first()

    def get_offset_sleep_time_per_day(self, sleep_col: str = None, based_on_diary: bool = False):
        """

        :param sleep_col:
        :param based_on_diary:
        :return: A Series indexed by experiment_day with total of minutes slept per day
        """

        if not based_on_diary and sleep_col is None:
            raise ValueError("Unable to calculate total sleep time."
                             " You have to specify a sleep column or set ``based_on_diary`` to True "
                             "(assuming you previously added a diary.")
        if based_on_diary:
            if self.diary is None:
                raise ValueError("Diary not found. Add a diary with ``add_diary``.")
            event = self.data[self.data[self.diary_offset] == True]
        else:
            if sleep_col not in self.data.keys():
                raise ValueError("Could not find sleep_col (%s). Aborting." % sleep_col)
            event = self.data[self.data[sleep_col] == True]

        return event.groupby(self.experiment_day_col)[self.time_col].last()

    def get_experiment_days(self):
        return list(self.data[self.get_experiment_day_col()].unique())


from unittest import TestCase

from hypnospy.data import RawProcessing

class TestRawProcessing(TestCase):

    def test_load_file(self):

        pp1 = RawProcessing()
        pp1.load_file("../data/examples_mesa/mesa-sample.csv",
                      collection_name="mesa",
                      col_for_datatime="linetime",
                      device_location="dw",
                      start_of_week="dayofweek",
                      col_for_pid="mesaid")
        nrows = pp1.data.shape[0]
        self.assertEqual(nrows, 23003)
        self.assertIn("linetime", pp1.data.keys())

        pp2 = RawProcessing()
        pp2.load_file("../data/examples_mesa/mesa-sample.csv",
                      collection_name="mesa",
                      col_for_datatime="linetime",
                      device_location="dw",
                      start_of_week=4,
                      pid=1)

        nrows = pp2.data.shape[0]
        self.assertEqual(nrows, 23003)
        self.assertIn("hyp_time", pp2.data.keys())

    def test_export_hypnospy(self):
        pass

    def test_run_nonwear(self):
        #self.fail()
        pass

    def test_calibrate_data(self):
        pass
        #self.fail()

    def test_obtain_pa_metrics(self):
        pass

    def test_extract_activity_index(self):
        pass

    def test_find_sleepwin(self):
        pass

    def test_obtain_data(self):
        pass



from unittest import TestCase

from hypnospy.data import RawProcessing
from hypnospy import Wearable
from hypnospy.analysis import TimeSeriesProcessing


class TestTimeSeriesProcessing(TestCase):

    def setUp(self):

        pp1 = RawProcessing()
        pp1.load_file("../data/examples_mesa/mesa-sample-day5-invalid5hours.csv",
                     # activitiy information
                     cols_for_activity=["activity"],
                     is_act_count=True,
                     # Datatime information
                     col_for_datatime="linetime",
                     device_location="dw",
                     start_of_week="dayofweek",
                     # Participant information
                     col_for_pid="mesaid")
        self.w_5day_invalid5hours = Wearable(pp1)


    def test_featurize(self):
        pass

    def test_day_night_split(self):
        pass

    def test_detect_sleep_boundaries(self):
        pass

    def test_fill_no_activity(self):
        pass

    def test_detect_non_wear(self):
        pass

    def test_configure_experiment_day(self):
        pass

    def test_check_valid_days(self):
        pass



from hypnospy import Wearable
from hypnospy import Experiment

import numpy as np


class SleepWakeAnalysis(object):

    def __init__(self, input: {Wearable, Experiment}):

        # TODO: we should have the option to only run sleep algorithms in the "hyp_night" period

        if type(input) is Wearable:
            self.wearables = [input]
        elif type(input) is Experiment:
            self.wearables = input.get_all_wearables()


    @staticmethod
    def __get_activity(wearable, activityIdx):

        if wearable is None:
            raise ValueError("Invalid wearable!")

        if activityIdx is not None:

            if activityIdx in wearable.data.keys():
                return activityIdx
            else:
                raise ValueError("Activity %s not found in the dataset. "
                                 "Options are: %s" % (activityIdx, ','.join(wearable.data.keys())))

        else:
            act = wearable.get_activity_col()

        return act

    def run_sleep_algorithm(self, algname, activityIdx=None, rescoring=False, on_sleep_interval=False, inplace=False, **kargs):
        """ Function that chooses sleep algorithm from a list of heuristic (non-data driven) approaches
            Function takes in algorithm (1) choice, (2) resolution (30s or 60s) currently available
            and (3) activity index method. Currently counts and ENMO (universally derivable across triaxial
            devices is available).
            
            * Function indicates if webster rescoring rules are applied
            
        Parameters
        ----------
        algname: str
            Heuristic method of choice
            Available methods are: 'ScrippsClinic', 'Sadeh', 'Oakley10', 'ColeKripke','Sazonov' 
            Default is 'ScrippsClinic'.
        Returns
        -------
        
        Citations (methods used and implemented)
        --------
        [1]Kripke, D. F., Hahn, E. K., ... & Kline, L. E. (2010). Wrist actigraphic scoring for
        sleep laboratory patients: algorithm development. Journal of sleep research, 19(4), 612-619.
        [2]Sadeh, A., Sharkey, M. & Carskadon, M. A. Activity-based sleep-wake identification: an empirical
        test of methodological issues. Sleep 17, 201–207 (1994).
        [3]Tonetti L, Pasquini F, Fabbri M, Belluzzi M, Natale V (2008) Comparison 
        of two different actigraphs with polysomnography in healthy young subjects. Chronobiol Int 25: 145–153
        [4]Cole, R. J., Kripke, D. F., Gruen, W., Mullaney, D. J. & Gillin, J. C. Automatic sleep/wake 
        identification from wrist activity. Sleep 15, 461–469 (1992).
        [5]Sazonov, E. et al. Activity-based sleep-wake identification in infants. Physiol. Meas. 25, 1291 (2004).
        [6]Tilmanne, J., Urbain, J., Kothare, M. V., Wouwer, A. V., & Kothare, S. V. (2009). Algorithms for sleep–wake
        identification using actigraphy: a comparative study and new results. Journal of sleep research, 18(1), 85-98.
        
        """
        results = {}

        for wearable in self.wearables:

            activityIdx = SleepWakeAnalysis.__get_activity(wearable, activityIdx)
            activity = wearable.data[activityIdx]

            # Apply Scripps
            if algname.lower() == "scrippsclinic":
                scaler = 0.204
                if "scaler" in kargs:
                    scaler = kargs["scaler"]
                result = self.__scripps_clinic_algorithm(activity, scaler)

            # Apply Sadeh
            elif algname.lower() == "sadeh":
                params = {"min_threshold": 0, "minNat": 50, "maxNat": 100, "window_past": 6, "window_nat": 11,
                          "window_centered": 11}
                for p in params:
                    if p in kargs:
                        params[p] = kargs[p]
                result = self.__sadeh_algorithm(activity, **params)

            # Apply Oakley
            elif algname.lower() == "oakley":
                threshold = 10
                if "threshold" in kargs:
                    threshold = kargs["threshold"]
                result = self.__oakley_algorithm(activity, threshold=threshold)

            # Apply Cole-Kripke
            elif algname.lower() in ["colekripke", "cole-kripke"]:
                result = self.__cole_kripke_algorithm(activity)

            # Apply Sazonov
            elif algname.lower() in ["sazonov", "saznov"]:
                result = self.__sazonov_algorithm(activity)

            else:
                raise ValueError("Algorithm %s is unknown." % (algname))

            if rescoring:
                result = self.webster_rescoring_rules(result)

            if inplace:
                wearable.data[algname] = result

            results[wearable.get_pid()] = result

        return results

    # %%
    def run_all_sleep_algorithms(self, activityIdx=None, rescoring=False, inplace=False):
        """ This function runs the algorithm of choice"
        """

        results = {}

        for wearable in self.wearables:
            activityIdx = SleepWakeAnalysis.__get_activity(wearable, activityIdx)
            activity = wearable.data[activityIdx]

            result = dict()
            result["ScrippsClinic"] = self.__scripps_clinic_algorithm(activity)
            result["Sadeh"] = self.__sadeh_algorithm(activity)
            result["Oakley10"] = self.__oakley_algorithm(activity, threshold=10)
            result["ColeKripke"] = self.__cole_kripke_algorithm(activity)
            result["Sazonov"] = self.__sazonov_algorithm(activity)

            if rescoring:
                result["RescoredScrippsClinic"] = self.webster_rescoring_rules(result["ScrippsClinic"])
                result["RescoredSadeh"] = self.webster_rescoring_rules(result["Sadeh"])
                result["RescoredOakley10"] = self.webster_rescoring_rules(result["Oakley10"])
                result["RescoredColeKripke"] = self.webster_rescoring_rules(result["ColeKripke"])
                result["RescoredSazonov"] = self.webster_rescoring_rules(result["Sazonov"])

            if inplace:
                for col in result.keys():
                    wearable.data[col] = result[col]

            results[wearable.get_pid()] = result

        return results


    # Scripps Clinic Algorithm Definition
    def __scripps_clinic_algorithm(self, activity, scaler=0.204):

        act_series = dict()
        act_series["_a0"] = activity.fillna(0.0)

        # Enrich the dataframe with temporary values
        for i in range(1, 11):
            act_series["_a-%d" % (i)] = activity.shift(i).fillna(0.0)
            act_series["_a+%d" % (i)] = activity.shift(-i).fillna(0.0)

        # Calculates Scripps clinic algorithm
        scripps = scaler * (0.0064 * act_series["_a-10"] + 0.0074 * act_series["_a-9"] +
                            0.0112 * act_series["_a-8"] + 0.0112 * act_series["_a-7"] +
                            0.0118 * act_series["_a-6"] + 0.0118 * act_series["_a-5"] +
                            0.0128 * act_series["_a-4"] + 0.0188 * act_series["_a-3"] +
                            0.0280 * act_series["_a-2"] + 0.0664 * act_series["_a-1"] +
                            0.0300 * act_series["_a0"] + 0.0112 * act_series["_a+1"] +
                            0.0100 * act_series["_a+2"])

        scripps.name = "ScrippsClinic"
        # Returns a series with binary values: 1 for sleep, 0 for awake
        return (scripps < 1.0).astype(int)

    # %%
    # Sadeh Algorithm
    def __sadeh_algorithm(self, activity, min_threshold=0, minNat=50, maxNat=100, window_past=6, window_nat=11,
                          window_centered=11):
        """
        Sadeh model for classifying sleep vs active
        """

        act = activity.copy()

        _mean = act.rolling(window=window_centered, center=True, min_periods=1).mean()
        _std = act.rolling(window=window_past, min_periods=1).std()
        _nat = ((act >= minNat) & (act <= maxNat)).rolling(window=window_nat, center=True, min_periods=1).sum()

        _LocAct = (act + 1.).apply(np.log)

        sadeh = (7.601 - 0.065 * _mean - 0.056 * _std - 0.0703 * _LocAct - 1.08 * _nat)
        sadeh.name = "Sadeh"

        # Returns a series with binary values: 1 for sleep, 0 for awake
        return (sadeh > min_threshold).astype(int)

    # %%
    # Oakley Algorithm
    def __oakley_algorithm(self, activity=None, threshold=80):
        """
        Oakley method to class sleep vs active/awake
        """
        act = activity.copy()

        act_series = {}

        act_series["_a0"] = act.fillna(0.0)
        for i in range(1, 5):
            act_series["_a-%d" % (i)] = act.shift(i).fillna(0.0)
            act_series["_a+%d" % (i)] = act.shift(-i).fillna(0.0)

        oakley = 0.04 * act_series["_a-4"] + 0.04 * act_series["_a-3"] + 0.20 * act_series["_a-2"] + \
                 0.20 * act_series["_a-1"] + 2.0 * act_series["_a0"] + 0.20 * act_series["_a+1"] + \
                 0.20 * act_series["_a-2"] + 0.04 * act_series["_a-3"] + 0.04 * act_series["_a-4"]
        oakley.name = "Oakley%d" % (threshold)

        return (oakley <= threshold).astype(int)

    # %%
    # Cole Kripke algorithm
    def __cole_kripke_algorithm(self, activity):
        """
        Cole-Kripke method to classify sleep vs awake
        """
        act_series = {}

        act_series["_A0"] = activity.fillna(0.0)
        for i in range(1, 5):
            act_series["_A-%d" % (i)] = activity.shift(i).fillna(0.0)
        for i in range(1, 3):
            act_series["_A+%d" % (i)] = activity.shift(-i).fillna(0.0)

        w_m4, w_m3, w_m2, w_m1, w_0, w_p1, w_p2 = [404, 598, 326, 441, 1408, 508, 350]
        p = 0.00001

        cole_kripke = p * (w_m4 * act_series["_A-4"] + w_m3 * act_series["_A-3"] +
                      w_m2 * act_series["_A-2"] + w_m1 * act_series["_A-1"] +
                      w_0 * act_series["_A0"] +
                      w_p1 * act_series["_A+1"] + w_p2 * act_series["_A+2"])

        cole_kripke.name = "Cole_Kripke"
        return (cole_kripke < 1.0).astype(int)

    # %%
    # Sazonov algorithm
    def __sazonov_algorithm(self, activity):
        """
        Sazonov formula as shown in the original paper
        """
        act_series = {}

        for w in range(1, 6):
            act_series["_w%d" % (w - 1)] = activity.rolling(window=w, min_periods=1).max()

        sazonov = 1.727 - 0.256 * act_series["_w0"] - 0.154 * act_series["_w1"] - \
                  0.136 * act_series["_w2"] - 0.140 * act_series["_w3"] - 0.176 * act_series["_w4"]
        sazonov.name = "Sazonov"
        return (sazonov >= 0.5).astype(int)


    # In the future include pre-trained ML/DL models here (optimize for features obtained)

    # Webster Rescoring Rules
    @staticmethod
    def webster_rescoring_rules(act, rescoring_rules="abcde"):

        if act.empty:
            return act

        haveAppliedAnyOtherRule = False

        if "a" in rescoring_rules or "A" in rescoring_rules:
            # After at least 4 minutes scored as wake, next minute scored as sleep is rescored wake
            # print "Processing rule A"
            maskA = act.shift(1).rolling(window=4, center=False,
                                         min_periods=1).sum() > 0  # avoid including actual period
            result = act.where(maskA, 0)
            haveAppliedAnyOtherRule = True

        if "b" in rescoring_rules or "B" in rescoring_rules:
            # After at least 10 minutes scored as wake, the next 3 minutes scored as sleep are rescored wake
            # print "Processing rule B"
            if haveAppliedAnyOtherRule == True:  # if this is true, I need to apply the next operation on the destination col
                act = result

            maskB = act.shift(1).rolling(window=10, center=False,
                                         min_periods=1).sum() > 0  # avoid including actual period
            result = act.where(maskB, 0).where(maskB.shift(1), 0).where(maskB.shift(2), 0)
            haveAppliedAnyOtherRule = True

        if "c" in rescoring_rules or "C" in rescoring_rules:
            # After at least 15 minutes scored as wake, the next 4 minutes scored as sleep are rescored as wake
            # print "Processing rule C"
            if haveAppliedAnyOtherRule == True:  # if this is true, I need to apply the next operation on the destination col
                act = result

            maskC = act.shift(1).rolling(window=15, center=False,
                                         min_periods=1).sum() > 0  # avoid including actual period
            result = act.where(maskC, 0).where(maskC.shift(1), 0).where(maskC.shift(2), 0).where(maskC.shift(3), 0)
            haveAppliedAnyOtherRule = True

        if "d" in rescoring_rules or "D" in rescoring_rules:
            # 6 minutes or less scored as sleep surroundeed by at least 10 minutes (before or after) scored as wake are rescored wake
            # print "Processing rule D"
            if haveAppliedAnyOtherRule == True:  # if this is true, I need to apply the next operation on the destination col
                act = result

            # First Part
            maskD1 = act.shift(1).rolling(window=10, center=False,
                                          min_periods=1).sum() > 0  # avoid including actual period
            tmpD1 = act.where(maskD1.shift(5), 0)
            haveAppliedAnyOtherRule = True

            # Second Part: sum the next 10 periods and replaces previous 6 in case they are all 0's
            maskD2 = act.shift(-10).rolling(window=10, center=False,
                                            min_periods=1).sum() > 0  # avoid including actual period
            tmpD2 = act.where(maskD2.shift(-5), 0)

            result = tmpD1 & tmpD2

        if "e" in rescoring_rules or "E" in rescoring_rules:
            # 10 minutes or less scored as sleep surrounded by at least 20 minutes (before or after) scored as wake are rescored wake
            # print "Processing rule E"
            if haveAppliedAnyOtherRule == True:  # if this is true, I need to apply the next operation on the destination col
                act = result

            # First Part
            maskE1 = act.shift(1).rolling(window=20, center=False,
                                          min_periods=1).sum() > 0  # avoid including actual period
            tmpE1 = act.where(maskE1.shift(9), 0)

            # Second Part: sum the next 10 periods and replaces previous 6 in case they are all 0's
            maskE2 = act.shift(-20).rolling(window=20, center=False,
                                            min_periods=1).sum() > 0  # avoid including actual period
            tmpE2 = act.where(maskE2.shift(-9), 0)

            result = tmpE1 & tmpE2

        return result



from .circadian_analysis import CircadianAnalysis
from .sleep_wake_analysis import SleepWakeAnalysis
from .non_wearing_detector import NonWearingDetector
from .phyisical_activity import PhysicalActivity
from .sleep_metrics import SleepMetrics
from .visualization import Viewer
from .sleep_boundary_detector import SleepBoudaryDetector
from .validator import Validator


from hypnospy import Wearable
from hypnospy import Experiment
import numpy as np
import warnings
from enum import IntFlag


class InvCode(IntFlag):
    # unfortunately the Flag type does not work properly with pandas 1.14
    FLAG_OKAY = 0
    FLAG_EPOCH_PA = 1
    FLAG_EPOCH_NON_WEARING = 2
    FLAG_EPOCH_NULL_VALUE = 4
    FLAG_DAY_SHORT_SLEEP = 8
    FLAG_DAY_LONG_SLEEP = 16
    FLAG_DAY_WITHOUT_DIARY = 32
    FLAG_DAY_NON_WEARING = 64
    FLAG_DAY_NOT_ENOUGH_VALID_EPOCHS = 128
    FLAG_DAY_NOT_ENOUGH_CONSECUTIVE_DAYS = 256

    @staticmethod
    def check_flag(int_value):
        return InvCode.FLAG_OKAY | int_value


class Validator(object):

    def __init__(self, input: {Wearable, Experiment}):
        """
        This module works on two levels: epochs and days.
        Methods started with ``flag_epoch_*`` will analyse epoch by epoch of the dataframe to decide whether the epoch should be marked as invalid or not.
        Methods started with ``flag_day_*`` will do the same at day level.

        If a day is flagged for removal, one can easily get rid of it with the method ``remove_flagged_days``.

        """
        self.wearables = {}

        if input is None:
            raise ValueError("Invalid value for input.")
        elif type(input) is Wearable:
            self.wearables[input.get_pid()] = input
        elif type(input) is Experiment:
            self.wearables = input.wearables

        self.invalid_col = "hyp_invalid"
        for wearable in self.wearables.values():
            wearable.data[self.invalid_col] = InvCode.FLAG_OKAY

    def flag_epoch_physical_activity_less_than(self, min_activity_threshold: int = 0):
        """
        Marks as invalid (InvCode.FLAG_PA) if physical activity is below ``min_activity_threshold``.

        :param min_activity_threshold: Integer threshold. Default: 0
        :return: None
        """
        for wearable in self.wearables.values():
            # Mark activity smaller than minimal
            wearable.data.loc[wearable.data[
                                  wearable.get_activity_col()] < min_activity_threshold, self.invalid_col] |= InvCode.FLAG_EPOCH_PA

    def flag_epoch_null_cols(self, col_list: list):
        """
        For each on of the columns in ``col_list``, this method marks as invalid (InvCode.FLAG_EPOCH_NULL_VALUE) if the value for an epoch is None/Null.

        :param col_list: List of columns to check for Null/None values.
        :return: None
        """

        for wearable in self.wearables.values():
            for col in col_list:
                if col not in wearable.data.keys():
                    raise KeyError("Col %s is not available for PID %s" % (col, wearable.get_pid()))

                wearable.data.loc[wearable.data[col].isnull(), self.invalid_col] |= InvCode.FLAG_EPOCH_NULL_VALUE

    def flag_epoch_nonwearing(self, wearing_col: str):
        """
        Marks as invalid (InvCode.FLAG_EPOCH_NON_WEARING) if epoch is considered non wearing.

        :param wearing_col: Column that was used by the NonWearingDetector Module.
        :return: None
        """
        for wearable in self.wearables.values():
            if wearing_col not in wearable.data.keys():
                raise KeyError(
                    "Column %s not found for wearable (pid=%s). Did you forget to run ``NonWearingDetector.detect_non_wear(...)``?" % (
                        wearing_col, wearable.get_pid()))

            wearable.data.loc[wearable.data[wearing_col] == False, self.invalid_col] |= InvCode.FLAG_EPOCH_NON_WEARING

    def flag_day_sleep_length_less_than(self, sleep_period_col: str, min_sleep_in_minutes: int):
        """
        Marks as invalid (InvCode.FLAG_SHORT_SLEEP) the whole day if the number of slept minutes is smaller than ``min_sleep_in_minutes``.
        This analysis is made based on the annotations of the ``sleep_period_col``.

        :param sleep_period_col: Binary sleep column containing: for each epoch, 1 is used for sleep and 0 for awake.
        :param min_sleep_in_minutes: Minimum required number of slept minutes per night.
        :return: None
        """

        for wearable in self.wearables.values():
            if sleep_period_col not in wearable.data.keys():
                warnings.warn("%s is not a valid entry in the data. "
                              "Maybe you need to run ``SleepBoudaryDetector.detect_sleep_boundaries(...)`` first. "
                              "Aborting...." % sleep_period_col)
                return

            sleep_time_per_day = wearable.get_total_sleep_time_per_day(sleep_period_col)
            days_with_problem = sleep_time_per_day < min_sleep_in_minutes
            days_with_problem = days_with_problem[days_with_problem[sleep_period_col] == True].index.values

            if days_with_problem.size > 0:
                wearable.data.loc[
                    wearable.data[wearable.get_experiment_day_col()].isin(
                        days_with_problem), self.invalid_col] |= InvCode.FLAG_DAY_SHORT_SLEEP

    def flag_day_sleep_length_more_than(self, sleep_period_col: str, max_sleep_in_minutes: int):
        """
        Marks as invalid (InvCode.FLAG_LONG_SLEEP) the whole day if the number of slept minutes is larger than ``max_sleep_in_minutes``.
        This analysis is made based on the annotations of the ``sleep_period_col``.

        :param sleep_period_col: Binary sleep column containing: for each epoch, 1 is used for sleep and 0 for awake.
        :param max_sleep_in_minutes: Maximum accepted number of slept minutes per night.
        :return: None
        """

        for wearable in self.wearables.values():
            if sleep_period_col not in wearable.data.keys():
                warnings.warn("%s is not a valid entry in the data. "
                              "Maybe you need to run ``SleepBoudaryDetector.detect_sleep_boundaries(...)`` first. "
                              "Aborting...." % sleep_period_col)
                return

            sleep_time_per_day = wearable.get_total_sleep_time_per_day(sleep_period_col)
            days_with_problem = sleep_time_per_day > max_sleep_in_minutes
            days_with_problem = days_with_problem[days_with_problem[sleep_period_col] == True].index.values

            if days_with_problem.size > 0:
                wearable.data.loc[
                    wearable.data[wearable.get_experiment_day_col()].isin(
                        days_with_problem), self.invalid_col] |= InvCode.FLAG_DAY_LONG_SLEEP

    def _flag_list_OR(self, wearable: Wearable, list_of_flags: list):
        """
        Internal method to evaluate if an InvCode was used or not.

        :param wearable: Wearable device
        :param list_of_flags: list of InvCode flags.
        :return: A pandas Series object with True if any of the InvCode were used in a given epoch.
        """

        result = wearable.data[self.invalid_col].apply(lambda x: list_of_flags[0] in InvCode.check_flag(x))
        for flag in list_of_flags[1:]:
            result |= wearable.data[self.invalid_col].apply(lambda x: flag in InvCode.check_flag(x))
        return result

    def flag_day_max_nonwearing(self, max_non_wear_minutes_per_day: int):
        """
        Marks as invalid the whole day (InvCode.FLAG_DAY_NON_WEARING) if non wearing time is greater than ``max_non_wear_minutes_per_day``.

        :param max_non_wear_minutes_per_day: Maximum tolerance for non wearing time per day (in minutes).
        :return: None
        """

        for wearable in self.wearables.values():
            epochs_in_minute = wearable.get_epochs_in_min()
            max_non_wear_epochs_per_day = max_non_wear_minutes_per_day * epochs_in_minute

            wearable.data["_tmp_flag_"] = self._flag_list_OR(wearable,
                                                             [InvCode.FLAG_EPOCH_PA, InvCode.FLAG_EPOCH_NON_WEARING,
                                                              InvCode.FLAG_EPOCH_NULL_VALUE])

            invalid_wearing = wearable.data.groupby([wearable.experiment_day_col])["_tmp_flag_"].transform(
                lambda x: x.sum()) >= max_non_wear_epochs_per_day
            wearable.data.loc[invalid_wearing, self.invalid_col] |= InvCode.FLAG_DAY_NON_WEARING
            del wearable.data["_tmp_flag_"]

    def flag_day_if_valid_epochs_smaller_than(self, valid_minutes_per_day: int):
        """
        Marks as invalid the whole day (InvCode.FLAG_DAY_NOT_ENOUGH_VALID_EPOCHS) if the number of valid minutes in a day is smaller than ``valid_minutes_per_day``.

        :param valid_minutes_per_day: Minimum minutes of valid (i.e., without any other flag) in a day.
        :return: None
        """

        for wearable in self.wearables.values():
            epochs_in_minute = wearable.get_epochs_in_min()
            valid_epochs_per_day = valid_minutes_per_day * epochs_in_minute

            wearable.data["_tmp_flag_"] = wearable.data[self.invalid_col] == InvCode.FLAG_OKAY

            invalid_epochs = wearable.data.groupby([wearable.experiment_day_col])["_tmp_flag_"].transform(
                lambda x: x.sum()) <= valid_epochs_per_day
            wearable.data.loc[invalid_epochs, self.invalid_col] |= InvCode.FLAG_DAY_NOT_ENOUGH_VALID_EPOCHS
            del wearable.data["_tmp_flag_"]

    def flag_day_without_diary(self):
        """
        Marks as invalid the whole day (InvCode.FLAG_DAY_WITHOUT_DIARY) if a diary entry is not found for that day.

        :return: None
        """

        for wearable in self.wearables.values():
            tst = wearable.get_total_sleep_time_per_day(based_on_diary=True)
            # Gets the experiment days with 0 total sleep time (i.e., no diary entry)
            invalid_days = set(tst[tst["hyp_diary_sleep"] == 0].index)
            # Flag them as invalid
            if len(invalid_days):
                wearable.data.loc[
                    wearable.data[wearable.get_experiment_day_col()].isin(
                        invalid_days), self.invalid_col] |= InvCode.FLAG_DAY_WITHOUT_DIARY

    def remove_flagged_days(self):
        """
        Fully removes from the wearable data the days that are flagged with problems.
        Only if all epochs in the day are flagged.

        :return: the total number of days removed
        """
        total_days_removed = 0
        for wearable in self.wearables.values():
            all_days = set(wearable.get_experiment_days())
            valid_days = set(self.get_valid_days(wearable.get_pid())[wearable.get_pid()])
            wearable.data = wearable.data[wearable.data[wearable.get_experiment_day_col()].isin(valid_days)].copy()

            # Get the length of the invalid_days without having to run get_invalid_daysL
            total_days_removed += len(all_days - valid_days)

        return total_days_removed

    def get_invalid_days(self, pid: str = None):
        """
        Returns a dictionary <pid> -> list of invalid days.

        :param pid: (optional) Wearable id. If pid is ``None'', this function returns the number of invalid days for all wearables.
        :return: list of invalid days in the dataset.

        """
        if pid is not None and pid in self.wearables:
            wlist = [self.wearables[pid]]
        else:
            wlist = self.wearables.values()

        r = {}
        for wearable in wlist:
            grp_days = wearable.data.groupby([wearable.get_experiment_day_col()])[self.invalid_col].all().reset_index()
            r[wearable.get_pid()] = set(
                grp_days[grp_days[self.invalid_col] == True][wearable.get_experiment_day_col()].unique())
        return r

    def get_valid_days(self, pid: str = None):
        """
        Returns a dictionary <pid> -> list of valid days.

        :param pid: (optional) Wearable id. If pid is ``None'', this function returns the number of valid days for all wearables.
        :return: list of valid days in the dataset.
        """

        if pid is not None and pid in self.wearables:
            wlist = [self.wearables[pid]]
        else:
            wlist = self.wearables.values()

        r = {}
        for wearable in wlist:
            invalid_days = self.get_invalid_days(wearable.get_pid())[wearable.get_pid()]
            all_days = set(wearable.data[wearable.get_experiment_day_col()].unique())
            r[wearable.get_pid()] = all_days - invalid_days
        return r

    def remove_wearable(self, pid: str):
        """
        Purges a wearable from the experiment.

        :param pid: Wearable id.
        :return: None
        """
        if pid in self.wearables:
            del self.wearables[pid]

    def remove_wearables_without_valid_days(self):
        """
        Fully removes any wearable if it does not have any valid data.

        :return: Total number of wearables removed.
        """
        mark_for_removal = []
        for wearable in self.wearables.values():
            valid_days = self.get_valid_days(wearable.get_pid())[wearable.get_pid()]
            if len(valid_days) == 0:
                mark_for_removal.append(wearable.get_pid())

        for pid in mark_for_removal:
            print("Removing wearable %s." % pid)
            self.remove_wearable(pid)

        return len(mark_for_removal)

    def flag_day_if_not_enough_consecutive_days(self, min_number_days: int):
        """
        In case the number of consecutive days is smaller than ``min_number_days``, we mark all as invalid.
        We also try to find any subset that has at least ``min_number_days``.

        :param min_number_days: minimum number of consecutive days
        :return: None
        """

        for wearable in self.wearables.values():

            days = self.get_valid_days(wearable.get_pid())[wearable.get_pid()]

            if len(days) == 0:
                return

            s = sorted(days)

            consecutive = 1
            last_value = s[0]
            saved_so_far = [last_value]
            okay = []

            for actual in s[1:]:
                if actual == last_value + 1:
                    consecutive += 1
                    last_value = actual
                    saved_so_far.append(last_value)

                else:
                    # Ops! We found a gap in the sequence.
                    # First we check if we already have enough days:
                    if len(saved_so_far) >= min_number_days:
                        okay.extend(saved_so_far)  # Cool! We have enough days.

                    else:  # Otherwise we start over
                        consecutive = 1
                        last_value = actual
                        saved_so_far = [last_value]

            if len(saved_so_far) >= min_number_days:
                okay.extend(saved_so_far)

            # In the okay set, we have all days that we can keep.
            new_invalid = set(days) - set(okay)
            if new_invalid:
                print("Flagging the following days as invalid for pid %s: %s" % (
                    wearable.get_pid(), ','.join(map(str, new_invalid))))

            wearable.data.loc[wearable.data[wearable.get_experiment_day_col()].isin(
                new_invalid), self.invalid_col] |= InvCode.FLAG_DAY_NOT_ENOUGH_CONSECUTIVE_DAYS

    def validation_report(self):
        """
        Generates a report from the actual state of flagged days.
        Days included in this report will be removed if use runs ''remove_flagged_days``.

        :return: None
        """

        day_related_checks = [InvCode.FLAG_DAY_SHORT_SLEEP, InvCode.FLAG_DAY_LONG_SLEEP, InvCode.FLAG_DAY_WITHOUT_DIARY,
                              InvCode.FLAG_DAY_NON_WEARING, InvCode.FLAG_DAY_NOT_ENOUGH_VALID_EPOCHS,
                              InvCode.FLAG_DAY_NOT_ENOUGH_CONSECUTIVE_DAYS]

        total_days = 0
        for check in day_related_checks:
            n_days_check_failed = 0
            for wearable in self.wearables.values():
                wearable.data["_tmp_flag_"] = self._flag_list_OR(wearable, [check])
                n_days_check_failed += wearable.data.groupby([wearable.experiment_day_col])["_tmp_flag_"].all().sum()
            total_days += n_days_check_failed
            print("Number of days removed due to %s: %d" % (check, n_days_check_failed))

        print("Total number of potential days to remove (may have overlaps): %d" % total_days)



import pandas as pd
import numpy as np


class RawProcessing(object):

    def __init__(self,
                 filename: str,
                 # Configuration for activity
                 cols_for_activity,
                 col_for_mets: object = None,
                 is_emno: bool = False,
                 is_act_count: bool = False,
                 # Datetime parameters
                 col_for_datetime: str = "time",
                 start_of_week: int = -1,
                 strftime: str = None,
                 # PID parameters
                 col_for_pid: str = None,
                 pid: int = -1,
                 # HR parameters
                 col_for_hr: str = None,
                 # Any additional data?
                 additional_data: object = None,
                 device_location: str = None,
                 ):

        """
        :param filename: input filepath
        :param device_location: where this device was located (options are: "bw", "hip", "dw", "ndw", "chest", "hp_ch", "hp_bw", "all")
        :param additional_data:
        """
        # self.possible_locations = ["bw", "hip", "dw", "ndw", "chest", "hp_ch", "hp_bw", "all"]
        self.device = None
        self.filename = filename
        self.device_location = device_location
        self.additional_data = additional_data

        self.internal_activity_cols = ["hyp_act_x", "hyp_act_y", "hyp_act_z"]
        self.internal_time_col = "hyp_time_col"
        self.internal_mets_col = None
        self.naxis = 0
        self.is_act_count = False
        self.is_emno = False
        self.pid = None

        self.data = self.__load_wearable_data(self.filename)
        self.__configure_activity(cols_for_activity, col_for_mets, is_emno, is_act_count)
        self.__configure_datetime(col_for_datetime, strftime, start_of_week)
        self.__configure_pid(col_for_pid, pid)
        self.__configure_hr(col_for_hr)

        # if device_location not in self.possible_locations:
        #    print("ERROR: Device location '%s' not implemented. Options are %s" % (device_location, ','.join(self.possible_locations)))

    def get_pid(self):
        return self.pid

    def set_time_col(self, new_name: str):
        if new_name is not None:
            self.internal_time_col = new_name

    def __configure_hr(self, col_for_hr):
        self.internal_hr_col = col_for_hr

    def __configure_activity(self, cols_for_activity, col_for_mets, is_emno, is_act_count):
        self.is_act_count = is_act_count
        self.is_emno = is_emno
        self.naxis = len(cols_for_activity)

        if self.naxis == 0:
            raise ValueError("Need at least one col to represent activity.")

        if len(cols_for_activity) > 3:
            raise ValueError("Current implementation allows up to 3 cols for physical activity.")

        if col_for_mets is not None:
            self.internal_mets_col = col_for_mets

        for i, col in enumerate(cols_for_activity):
            if col not in self.data.keys():
                raise ValueError(
                    "Col %s not detected in the dataset. Possibilities are %s" % (col, ','.join(self.data.keys())))
            # If col exists, we save it with our internal name.
            # Note that in case only one axis is available, it will be 'hyp_act_x'.
            self.data[self.internal_activity_cols[i]] = self.data[col]

    def __configure_pid(self, col_for_pid: str, pid: int):
        if col_for_pid is None and pid == -1:
            raise ValueError("Either pid or col_for_pid need to have a valid value.")

        if pid != -1:
            self.pid = pid

        elif col_for_pid is not None:
            if col_for_pid not in self.data.keys():
                raise ValueError("Column %s is not in the dataframe." % (col_for_pid))

            pid = self.data.iloc[0][col_for_pid]
            self.pid = pid

    def __configure_datetime(self, col_for_datetime, strftime, start_of_week):
        if strftime is None and start_of_week is None:
            raise ValueError("Either strftime or start_of_week need to have a valid value.")

        # We need to figure out when the data started being collected.
        # Some datasets like MESA and HCHS from sleepdata.org do not have date information, unfortunately
        if strftime is None or (strftime is not None and strftime.find("%d") == -1):  # Could not find a day
            # Check if we can extract the start_of_week:
            starting_day_of_week = 1
            if type(start_of_week) == str:  # ColName from which we will extract the start_of_week
                if start_of_week not in self.data:
                    raise ValueError("%s is not a column in the dataframe" % (start_of_week))

                starting_day_of_week = self.data.iloc[0][start_of_week]

            elif type(start_of_week) == int:
                starting_day_of_week = start_of_week

            self.__datetime_without_date(col_for_datetime, starting_day_of_week)

        # We know when the week started because we have strftime well defined:
        else:
            self.data[self.internal_time_col] = pd.to_datetime(self.data[col_for_datetime], format=strftime)

    def __datetime_without_date(self, col_for_datatime, starting_day_of_week):
        """
        If we blindly use something like

            df["linetime"] = pd.to_datetime(df["linetime"])

        we will have all the days in the dataset set as the current day.
        That is not what we want.

        Instead, we pick a starting day (January 1, 2017 - *Sunday*) and modify the
        rest of the data according to the day of th week the experiment started.

        This way, if an experiment started on a Tuesday (day=3), the actigraphy
        data for this person would start on January 3, 2017.

        :param col_for_datatime: col name in the dataframe for the datetime
        :param starting_day_of_week: 0 = Sunday, 1 = Monday, ... 6 = Saturday
        """

        # TODO: this procedure to find the freq might be too specific and work only for mesa and latinos
        freq = abs(int(self.data[col_for_datatime].iloc[1][-2:]) - int(self.data[col_for_datatime].iloc[0][-2:]))
        ndays = int(np.ceil(self.data.shape[0] / (24 * (60 / freq) * 60)))
        firstTime, lastTime = self.data.iloc[0][col_for_datatime], self.data.iloc[-1][col_for_datatime]

        for n in range(-1, 2):
            times = pd.date_range(start="1-%d-2017 %s" % (starting_day_of_week, firstTime),
                                  end="1-%d-2017 %s" % (starting_day_of_week + ndays + n, lastTime),
                                  freq="{}s".format(freq))

            if times.shape[0] == self.data.shape[0]:
                self.data[self.internal_time_col] = times
                break
        else:
            raise ValueError("Could not find correct range for dataframe. "
                             "Please check if parameter ``datatime_col'' (=%s) is correct and has all its entries valid." % col_for_datatime)

    def export_hypnospy(self, filename):
        """

        :param filename:
        :return:
        """

        # TODO: find a better way to save these fields to file
        self.data.to_hdf(filename, key='data', mode='w')
        s = pd.Series([self.pid, self.internal_time_col,
                       self.internal_activity_cols[0:self.naxis],
                       self.internal_mets_col,
                       self.is_act_count, self.is_emno,
                       self.device_location, self.additional_data])
        s.to_hdf(filename, key='other')

        # mydict = dict(data=self.data,
        #               location=self.device_location,
        #               additional=self.additional_data)
        #
        # with h5py.File(filename, 'w') as hf:
        #     grp = hf.create_group('alist')
        #     for k, v in mydict.items():
        #         grp.create_dataset(k, data=v)

        # hf.create_dataset('data', data=mydata)

        # hf.create_dataset('location', data=)
        # hf.create_dataset('additional_data', data=self.additional_data)

        print("Saved file %s." % (filename))

    def __load_wearable_data(self, filename):
        """ Obtain device type
        Used to decide which way to parse the data from input file

        :return: Wearable type (Axivity, GeneActiv, Actigraph, Actiwatch and Apple Watch currently supported)
        """
        f = filename.lower()
        if f.endswith('.cwa') or f.endswith('.cwa.gz') or f.endswith('CWA'):
            return self.__process_axivity(filename)
        elif f.endswith('.bin'):
            return self.__process_geneactiv(filename)
        elif f.endswith('.dat'):
            # here ask for which device: ={ "applewatch":"Apple Watch","apple" [...] "actiwatch"[...]
            # ask if the data is raw or not--> if not--> move down the pecking order
            return self.__process_actigraph(filename)
        elif f.endswith('.csv') or f.endswith('.csv.gz'):
            return self.__process_csv(filename)
        else:
            print("ERROR: Wearable format not supported for file: " + filename)

    def __process_csv(self, csvfile):
        return pd.read_csv(csvfile)

    def __process_axivity(self, cwaFile):
        """ Use PAMPRO here
        """
        pass

    def __process_actigraph(self, datFile):
        """ Use PAMPRO here
        """
        pass

    def __process_geneactiv(self, datFile):
        """ Use PAMPRO here
        """
        pass

    def run_nonwear():
        """ Use PAMPRO autocalibration
        """
        pass

    def calibrate_data():
        """ Use PAMPRO autocalibration
        """
        pass

    def obtain_PA_metrics():
        """ Use PAMPRO for derivation of VM, VMHPF, ENMO, Pitch, Roll, Yaw
              Include proxy for MVPA, PA
        """
        # run this if class isn't count based
        # returns something that can be then summarized
        pass


class ActiwatchSleepData(RawProcessing):

    def __init__(self, filename, device_location=None, col_for_datetime="time", col_for_pid="pid"):
        super().__init__(filename, device_location=device_location,
                         cols_for_activity=["activity"],
                         is_act_count=True,
                         # Datatime information
                         col_for_datetime=col_for_datetime,
                         start_of_week="dayofweek",
                         # Participant information
                         col_for_pid=col_for_pid,
                         )
        self.device = "actigraphy"
        self.data["hyp_annotation"] = self.data["interval"].isin(["REST", "REST-S"])


# TODO: missing Actiheart (Fenland, BBVS), Axivity (BBVS, Biobank)

class MESAPreProcessing(RawProcessing):

    def __init__(self, filename, device_location=None, col_for_datetime="linetime", col_for_pid="mesaid"):
        super().__init__(filename, device_location=device_location,
                         cols_for_activity=["activity"],
                         is_act_count=True,
                         # Datatime information
                         col_for_datetime=col_for_datetime,
                         start_of_week="dayofweek",
                         # Participant information
                         col_for_pid=col_for_pid,
                         )
        self.device = "actigraphy"
        self.data["hyp_annotation"] = self.data["interval"].isin(["REST", "REST-S"])


class MMASHPreProcessing(RawProcessing):

    def __init__(self, filename, device_location=None, col_for_datetime="time", col_for_pid="pid",
                 col_for_hr="HR", cols_for_activity=["Axis1", "Axis2", "Axis3"], strftime="%Y-%b-%d %H:%M:%S"):
        super().__init__(filename, device_location=device_location,
                         # Activity information
                         cols_for_activity=cols_for_activity,
                         # Datatime information
                         col_for_datetime=col_for_datetime,
                         strftime=strftime,
                         # Participant information
                         col_for_pid=col_for_pid,
                         # HR information:
                         col_for_hr=col_for_hr,
                         )
        self.device = "actigraphy"



import pandas as pd

class Diary(object):

    def __init__(self):
        """

        :param input: Either a path to a PreProcessing file saved with ``export_hyp`` or a PreProcessing object
        """
        self.data = None

    def from_file(self, file_path):
        self.data = pd.read_csv(file_path)
        if "pid" not in self.data.keys():
            # TODO: decide if we can allow the user to specify it.
            raise KeyError("Diary needs to have a 'pid' column.")

        self.data["pid"] = self.data["pid"].astype(str)
        return self

    def from_dataframe(self, dataframe):
        self.data = dataframe
        return self


from hypnospy import Wearable
from hypnospy import Experiment
from hypnospy import misc
import numpy as np
import pandas as pd
from datetime import timedelta
import warnings


class SleepBoudaryDetector(object):

    def __init__(self, input: {Wearable, Experiment}):
        """ Here we need to load the data and determine:
            potentially by fetching from class wearable
            (1) what type of file is it
            (2) is it multimodal
            (3) length/type- night only/ full
            (4) sampling rate
            """

        if type(input) is Wearable:
            self.wearables = [input]
        elif type(input) is Experiment:
            self.wearables = input.get_all_wearables()

    @staticmethod
    def __create_threshold_col_based_on_time(df, time_col: str, hr_col: str, start_time: int, end_time: int,
                                             quantile: float, rolling_win_in_minutes: int,
                                             sleep_only_in_sleep_search_window: bool):

        df_time = df.set_index(time_col).copy()

        # This will get the index of everything outside <start_time, end_time>
        idx = df_time.between_time('%02d:00' % end_time,
                                   '%02d:00' % start_time,
                                   include_start=False,
                                   include_end=False).index
        # We set the hr_col to nan for the time outside the search win in order to find the quantile below ignoring nans
        df_time.loc[idx, hr_col] = np.nan

        quantiles_per_day = df_time[hr_col].resample('24H', base=start_time).quantile(quantile).dropna()
        df_time["hyp_sleep"] = quantiles_per_day
        if quantiles_per_day.index[0] < df_time.index[0]:
            df_time.loc[df_time.index[0], "hyp_sleep"] = quantiles_per_day.iloc[0]

        # We fill the nans in the df_time and copy the result back to the original df
        df_time["hyp_sleep"] = df_time["hyp_sleep"].fillna(method='ffill').fillna(method='bfill')

        # binarize_by_hr_threshold
        df_time["hyp_sleep_bin"] = np.where((df_time[hr_col] - df_time["hyp_sleep"]) > 0, 0, 1)
        df_time["hyp_sleep_bin"] = df_time["hyp_sleep_bin"].rolling(window=rolling_win_in_minutes).median().fillna(
            method='bfill')

        if sleep_only_in_sleep_search_window:
            #  Ignore all sleep candidate period outsite win
            df_time.loc[idx, "hyp_sleep_bin"] = 0

        seq_length, seq_id = misc.get_consecutive_serie(df_time, "hyp_sleep_bin")

        return df_time["hyp_sleep"].values, df_time["hyp_sleep_bin"].values, seq_length.values, seq_id.values

    @staticmethod
    def __sleep_boundaries_with_hr(wearable: Wearable, output_col: str, quantile: float = 0.4,
                                   volarity_threshold: int = 5, rolling_win_in_minutes: int = 5,
                                   sleep_search_window: tuple = (20, 12), min_window_length_in_minutes: int = 40,
                                   volatility_window_in_minutes: int = 10, merge_blocks_gap_time_in_min: int = 240,
                                   sleep_only_in_sleep_search_window: bool = False,
                                   only_largest_sleep_period: bool = False):

        if wearable.hr_col is None:
            raise AttributeError("HR is not available for PID %s." % (wearable.get_pid()))

        rolling_win_in_minutes = int(rolling_win_in_minutes * wearable.get_epochs_in_min())
        min_window_length_in_minutes = int(min_window_length_in_minutes * wearable.get_epochs_in_min())
        volatility_window_in_minutes = int(volatility_window_in_minutes * wearable.get_epochs_in_min())

        df = wearable.data.copy()

        df["hyp_sleep"], df["hyp_sleep_bin"], df["hyp_seq_length"], df[
            "hyp_seq_id"] = SleepBoudaryDetector.__create_threshold_col_based_on_time(wearable.data, wearable.time_col,
                                                                                      wearable.hr_col,
                                                                                      sleep_search_window[0],
                                                                                      sleep_search_window[1],
                                                                                      quantile,
                                                                                      rolling_win_in_minutes,
                                                                                      sleep_only_in_sleep_search_window)

        df['hyp_sleep_candidate'] = ((df["hyp_sleep_bin"] == 1.0) & (
                df['hyp_seq_length'] > min_window_length_in_minutes)).astype(int)

        df["hyp_sleep_vard"] = df[wearable.hr_col].rolling(volatility_window_in_minutes,
                                                           center=True).std().fillna(0)

        df["hyp_seq_length"], df["hyp_seq_id"] = misc.get_consecutive_serie(df, "hyp_sleep_candidate")

        # Merge two sleep segments if their gap is smaller than X min (interval per day):
        wearable.data = df
        saved_hour_start_day = wearable.hour_start_experiment
        wearable.change_start_hour_for_experiment_day(sleep_search_window[0])
        grps = wearable.data.groupby(wearable.experiment_day_col)
        tmp_df = []
        for grp_id, grp_df in grps:
            gdf = grp_df.copy()
            gdf["hyp_sleep_candidate"], gdf["hyp_seq_id"], gdf["hyp_seq_length"] = misc.merge_windows(gdf,
                                                                                                      wearable.time_col,
                                                                                                      "hyp_sleep_candidate",
                                                                                                      tolerance_minutes=merge_blocks_gap_time_in_min)

            tmp_df.append(gdf)
        wearable.data = pd.concat(tmp_df)
        wearable.change_start_hour_for_experiment_day(saved_hour_start_day)

        df = wearable.data.set_index(wearable.time_col)
        new_sleep_segments = df[df["hyp_sleep_candidate"] == 1]["hyp_seq_id"].unique()

        # Check if we can modify the sleep onset/offset
        for sleep_seg_id in new_sleep_segments:
            actual_seg = df[df["hyp_seq_id"] == sleep_seg_id]

            if actual_seg.shape[0] == 0:
                continue

            start_time = actual_seg.index[0]
            end_time = actual_seg.index[-1]

            look_sleep_onset = df[start_time - timedelta(hours=4): start_time + timedelta(minutes=60)]
            look_sleep_offset = df[end_time - timedelta(minutes=1): end_time + timedelta(minutes=120)]

            new_sleep_onset = look_sleep_onset[look_sleep_onset["hyp_sleep_vard"] > volarity_threshold]
            new_sleep_offset = look_sleep_offset[look_sleep_offset["hyp_sleep_vard"] > volarity_threshold]

            new_start = new_sleep_onset.index[-1] if not new_sleep_onset.empty else start_time
            new_end = new_sleep_offset.index[0] if not new_sleep_offset.empty else end_time

            df.loc[new_start:new_end, "hyp_seq_id"] = sleep_seg_id
            # df.loc[new_start:new_end, "hyp_seq_length"] = df.loc[new_start:new_end].shape[0]
            df.loc[new_start:new_end, "hyp_sleep_candidate"] = 1

        # Need to reorganize the sequences.
        df["hyp_seq_length"], df["hyp_seq_id"] = misc.get_consecutive_serie(df, "hyp_sleep_candidate")

        # new_sleep_segments = df[df[col_win_night + '_sleep_candidate'] == 1][col_win_night + '_grpid'].unique()
        wearable.data = df.reset_index()

        if only_largest_sleep_period:  # If true, we keep only one sleep period per night.

            saved_hour_start_day = wearable.hour_start_experiment
            wearable.change_start_hour_for_experiment_day(sleep_search_window[0])

            grps = wearable.data.groupby(wearable.experiment_day_col)
            tmp_df = []
            for grp_id, grp_df in grps:
                gdf = grp_df.copy()
                gdf["hyp_seq_length"], gdf["hyp_seq_id"] = misc.get_consecutive_serie(gdf, "hyp_sleep_candidate")
                df_out = misc.find_largest_sequence(gdf, "hyp_sleep_candidate", output_col).replace(-1, False)
                tmp_df.append(df_out)
            wearable.data[output_col] = pd.concat(tmp_df)
            wearable.change_start_hour_for_experiment_day(saved_hour_start_day)
        else:
            # Save final output
            wearable.data[output_col] = False
            wearable.data.loc[wearable.data[(wearable.data["hyp_sleep_candidate"] == 1)].index, output_col] = True

        # Clean up!
        wearable.data.drop(
            columns=["hyp_sleep", "hyp_sleep_candidate", "hyp_seq_id",
                     "hyp_sleep_bin",
                     "hyp_sleep_vard", "hyp_seq_length"], inplace=True)

    @staticmethod
    def __sleep_boundaries_with_annotations(wearable, output_col, annotation_col, hour_to_start_search=18,
                                            merge_tolerance_in_minutes=20, only_largest_sleep_period=True):

        if annotation_col not in wearable.data.keys():
            raise KeyError("Col %s is not a valid for pid %s" % (annotation_col, wearable.get_pid()))

        saved_hour_start_day = wearable.hour_start_experiment
        wearable.change_start_hour_for_experiment_day(hour_to_start_search)

        wearable.data["hyp_sleep_candidate"] = wearable.data[annotation_col].copy()

        # Annotates the sequences of sleep_candidate
        wearable.data["hyp_seq_length"], wearable.data["hyp_seq_id"] = misc.get_consecutive_serie(wearable.data,
                                                                                                  "hyp_sleep_candidate")

        wearable.data["hyp_sleep_candidate"], wearable.data["hyp_seq_id"], wearable.data[
            "hyp_seq_length"] = misc.merge_windows(wearable.data, wearable.time_col, "hyp_sleep_candidate",
                                                   merge_tolerance_in_minutes)

        if only_largest_sleep_period:
            grps = wearable.data.groupby(wearable.experiment_day_col)
            tmp_df = []
            for grp_id, grp_df in grps:
                gdf = grp_df.copy()
                gdf["hyp_seq_length"], gdf["hyp_seq_id"] = misc.get_consecutive_serie(gdf, "hyp_sleep_candidate")
                df_out = misc.find_largest_sequence(gdf, "hyp_sleep_candidate", output_col).replace(-1, False)
                tmp_df.append(df_out)
            wearable.data[output_col] = pd.concat(tmp_df)
        else:
            wearable.data[output_col] = False
            wearable.data.loc[wearable.data[(wearable.data["hyp_sleep_candidate"] == 1)].index, output_col] = True

        del wearable.data["hyp_seq_id"]
        del wearable.data["hyp_seq_length"]
        del wearable.data["hyp_sleep_candidate"]

        wearable.change_start_hour_for_experiment_day(saved_hour_start_day)

    @staticmethod
    def __sleep_boundaries_with_angle_change_algorithm(wearable: Wearable, output_col: str,
                                                       start_hour: int = 15,
                                                       cols: list = [],
                                                       use_triaxial_activity=False,
                                                       q_sleep: float = 0.1,
                                                       minimum_len_in_minutes: int = 30,
                                                       merge_tolerance_in_minutes: int = 180,
                                                       factor: int = 15,
                                                       operator: str = "or",  # Either 'or' or 'and'
                                                       only_largest_sleep_period: bool = False
                                                       ):

        df_time = wearable.data.copy()
        df_time = df_time.set_index(wearable.time_col)

        five_min = int(5 * wearable.get_epochs_in_min())
        minimum_len_in_minutes = int(minimum_len_in_minutes * wearable.get_epochs_in_min())

        if use_triaxial_activity:
            # Step 1:
            df_time["hyp_rolling_x"] = df_time["hyp_act_x"].rolling("5s").median().fillna(0.0)
            df_time["hyp_rolling_y"] = df_time["hyp_act_y"].rolling("5s").median().fillna(0.0)
            df_time["hyp_rolling_z"] = df_time["hyp_act_z"].rolling("5s").median().fillna(0.0)

            df_time["hyp_act_z"].rolling(five_min).median().fillna(0.0)

            df_time["hyp_angle_z"] = (np.arctan(
                df_time["hyp_rolling_z"] / ((df_time['hyp_rolling_y'] ** 2 + df_time['hyp_rolling_x'] ** 2) ** (
                        1 / 2)))) * 180 / np.pi
            # Step 2:
            df_time["hyp_angle_z"] = df_time["hyp_angle_z"].fillna(0.0)
            # Step 3:
            df_time["hyp_angle_z"] = df_time["hyp_angle_z"].rolling("5s").mean().fillna(0.0)

            cols += ["hyp_angle_z"]

        if operator == "or":
            df_time["hyp_sleep_candidate"] = False
        else:
            df_time["hyp_sleep_candidate"] = True

        for col in cols:
            # Paper's Step 4
            df_time["hyp_" + col + '_diff'] = df_time[col].diff().abs()
            # Paper's Step 5
            df_time["hyp_" + col + '_5mm'] = df_time["hyp_" + col + '_diff'].rolling(five_min).median().fillna(0.0)
            # Paper's Step 6
            quantiles_per_day = df_time["hyp_" + col + '_5mm'].resample('24H', base=start_hour).quantile(
                q_sleep).dropna()
            # print(quantiles_per_day)

            df_time["hyp_" + col + '_10pct'] = quantiles_per_day
            if quantiles_per_day.index[0] < df_time.index[0]:
                df_time.loc[df_time.index[0], "hyp_" + col + '_10pct'] = quantiles_per_day.iloc[0]

            df_time["hyp_" + col + '_10pct'] = df_time["hyp_" + col + '_10pct'].fillna(method='ffill').fillna(
                method='bfill')

            df_time["hyp_" + col + '_bin'] = np.where(
                (df_time["hyp_" + col + '_5mm'] - (df_time["hyp_" + col + '_10pct'] * factor)) > 0, 0, 1)
            df_time["hyp_" + col + '_len'], _ = misc.get_consecutive_serie(df_time, "hyp_" + col + '_bin')

            # Paper's Step 7
            if operator == "or":
                df_time["hyp_sleep_candidate"] = df_time["hyp_sleep_candidate"] | (
                            (df_time["hyp_" + col + '_bin'] == 1.0) & (
                            df_time["hyp_" + col + '_len'] > minimum_len_in_minutes))
            else:
                df_time["hyp_sleep_candidate"] = df_time["hyp_sleep_candidate"] & \
                                                 ((df_time["hyp_" + col + '_bin'] == 1.0)
                                                  & (df_time["hyp_" + col + '_len'] > minimum_len_in_minutes))

        # Gets the largest sleep_candidate per night
        wearable.data = df_time.reset_index()
        # wearable.data[output_col] = wearable.data["hyp_sleep_candidate"]

        wearable.data["hyp_seq_length"], wearable.data["hyp_seq_id"] = misc.get_consecutive_serie(wearable.data,
                                                                                                  "hyp_sleep_candidate")
        # Paper's Step 8
        wearable.data["hyp_sleep_candidate"], wearable.data["hyp_seq_id"], wearable.data[
            "hyp_seq_length"] = misc.merge_windows(wearable.data, wearable.time_col, "hyp_sleep_candidate",
                                                   merge_tolerance_in_minutes)

        # Paper's Step 9
        if only_largest_sleep_period:  # If true, we keep only one sleep period per night.
            saved_hour_start_day = wearable.hour_start_experiment
            wearable.change_start_hour_for_experiment_day(start_hour)
            grps = wearable.data.groupby(wearable.experiment_day_col)
            tmp_df = []
            for grp_id, grp_df in grps:
                gdf = grp_df.copy()
                gdf["hyp_seq_length"], gdf["hyp_seq_id"] = misc.get_consecutive_serie(gdf, "hyp_sleep_candidate")
                df_out = misc.find_largest_sequence(gdf, "hyp_sleep_candidate", output_col).replace(-1, False)
                tmp_df.append(df_out)
            wearable.data[output_col] = pd.concat(tmp_df)
            wearable.change_start_hour_for_experiment_day(saved_hour_start_day)
        else:
            # Save final output
            wearable.data[output_col] = False
            wearable.data.loc[wearable.data[(wearable.data["hyp_sleep_candidate"] == 1)].index, output_col] = True

        # Cleaning up...
        cols_to_drop = ["hyp_sleep_candidate", "hyp_seq_length", "hyp_seq_id"]
        for col in cols:
            cols_to_drop.append("hyp_" + col + '_diff')
            cols_to_drop.append("hyp_" + col + '_5mm')
            cols_to_drop.append("hyp_" + col + '_10pct')
            cols_to_drop.append("hyp_" + col + '_len')

        wearable.data.drop(columns=cols_to_drop, inplace=True)

    def detect_sleep_boundaries(self, strategy: str, output_col: str = "hyp_sleep_period",
                                annotation_hour_to_start_search: int = 18, annotation_col: str = None,
                                annotation_merge_tolerance_in_minutes: int = 20,
                                annotation_only_largest_sleep_period: bool = True,
                                hr_quantile: float = 0.4, hr_volarity_threshold: int = 5,
                                hr_rolling_win_in_minutes: int = 5, hr_sleep_search_window: tuple = (20, 12),
                                hr_min_window_length_in_minutes: int = 40, hr_volatility_window_in_minutes: int = 10,
                                hr_merge_blocks_gap_time_in_min: int = 240,
                                hr_sleep_only_in_sleep_search_window: bool = False,
                                hr_only_largest_sleep_period: bool = False,
                                angle_cols: list = [],
                                angle_use_triaxial_activity: bool = False, angle_start_hour: int = 15,
                                angle_quantile: float = 0.1, angle_minimum_len_in_minutes: int = 30,
                                angle_merge_tolerance_in_minutes: int = 180,
                                angle_only_largest_sleep_period: bool = True):
        """
            Detected the sleep boundaries.

            param
            -----

            strategy: "annotation", "hr", "angle"

            Creates a new col (output_col, default: 'hyp_sleep_period')
            which has value 1 if inside the sleep boundaries and 0 otherwise.

        """
        # (1) HR sleeping window approach here (SLEEP 2020 paper)
        # (2) expert annotations or PSG
        # (3) Van Hees heuristic method
        # Missing: Crespo (periods of innactivity)

        for wearable in self.wearables:
            # print(wearable.get_pid())

            if wearable.data.shape[0] == 0:
                wearable.data[output_col] = np.nan
                warnings.warn("No data for PID %s. Skipping it." % wearable.get_pid())
                continue

            if strategy == "annotation":
                self.__sleep_boundaries_with_annotations(wearable, output_col, annotation_col,
                                                         annotation_hour_to_start_search,
                                                         annotation_merge_tolerance_in_minutes,
                                                         annotation_only_largest_sleep_period)
            elif strategy == "hr":
                self.__sleep_boundaries_with_hr(wearable, output_col, hr_quantile, hr_volarity_threshold,
                                                hr_rolling_win_in_minutes, hr_sleep_search_window,
                                                hr_min_window_length_in_minutes, hr_volatility_window_in_minutes,
                                                hr_merge_blocks_gap_time_in_min, hr_sleep_only_in_sleep_search_window,
                                                hr_only_largest_sleep_period)

            elif strategy.lower() in ["adapted_van_hees", "angle"]:
                self.__sleep_boundaries_with_angle_change_algorithm(wearable,
                                                                    output_col=output_col,
                                                                    cols=angle_cols,
                                                                    use_triaxial_activity=angle_use_triaxial_activity,
                                                                    start_hour=angle_start_hour,
                                                                    q_sleep=angle_quantile,
                                                                    minimum_len_in_minutes=angle_minimum_len_in_minutes,
                                                                    merge_tolerance_in_minutes=angle_merge_tolerance_in_minutes,
                                                                    only_largest_sleep_period=angle_only_largest_sleep_period,
                                                                    )



            else:
                warnings.warn("Strategy %s is not yet implemented" % (strategy))



import warnings
from datetime import timedelta


def get_consecutive_serie(df_in, col):
    df = df_in.copy()

    df["_lenght"] = 0
    df["_seq_id"] = df.groupby((df[col] != df[col].shift(1).fillna(False)).cumsum()).ngroup()
    df["_lenght"] = df[["_lenght", "_seq_id"]].groupby("_seq_id").transform("count")

    return df["_lenght"], df["_seq_id"]


def find_largest_sequence(df_orig, candidate, output_col, seq_length_col="hyp_seq_length", seq_id_col="hyp_seq_id"):
    df = df_orig.copy()
    df[output_col] = False

    df_candidates = df[df[candidate] == True]

    if df_candidates.empty:
        warnings.warn("Day has no valid elements for column %s." % candidate)
        df[output_col] = -1
        return df[output_col]

    #print(df_candidates[seq_id_col].unique(), df_candidates[seq_length_col].unique())
    # Mark the largest period as "sleep_period"
    largest_seqid = df_candidates.iloc[df_candidates[seq_length_col].argmax()][seq_id_col]
    largest = df_candidates[df_candidates[seq_id_col] == largest_seqid]
    df.loc[largest.index, output_col] = True

    return df[output_col]

def merge_windows(df_orig, time_col, sleep_candidate_col, tolerance_minutes=20):
    """

    :param df_orig:
    :param time_col:
    :param sleep_candidate_col:
    :param tolerance_minutes: Use negative to force merging first and last boards
    :return:
    """

    df = df_orig.copy()
    saved_index = df.reset_index()["index"]
    df = df.set_index(time_col)

    df_candidates = df[df[sleep_candidate_col] == True]

    if df_candidates.shape[0] == 0:
        warnings.warn("Day has no sleep period!")
        return df[sleep_candidate_col], df["hyp_seq_id"], df["hyp_seq_length"]

    if tolerance_minutes > 0:

        # Get the list of all sleep candidates
        all_seq_ids = sorted(df_candidates["hyp_seq_id"].unique())  # What are the possible seq_ids?

        actual_sleep_seg_id = all_seq_ids[0]

        for next_sleep_seg_id in all_seq_ids[1:]:

            actual_segment = df[df["hyp_seq_id"] == actual_sleep_seg_id]
            start_time_actual_seg = actual_segment.index[0]
            end_time_actual_seg = actual_segment.index[-1]

            next_segment = df[df["hyp_seq_id"] == next_sleep_seg_id]
            start_time_next_segment = next_segment.index[0]
            end_time_next_segment = next_segment.index[-1]

            if start_time_next_segment - end_time_actual_seg <= timedelta(minutes=tolerance_minutes):
                # Merges two sleep block
                df.loc[start_time_actual_seg:end_time_next_segment, "hyp_seq_id"] = actual_sleep_seg_id
                df.loc[start_time_actual_seg:end_time_next_segment, "hyp_seq_length"] = df.loc[start_time_actual_seg:end_time_next_segment].shape[0]
                df.loc[start_time_actual_seg:end_time_next_segment, "hyp_sleep_candidate"] = True
            else:
                actual_sleep_seg_id = next_sleep_seg_id

    else:
        df.loc[df_candidates.index[0]:df_candidates.index[-1], "hyp_sleep_candidate"] = True
        df.loc[df_candidates.index[0]:df_candidates.index[-1], "hyp_seq_length"] = df.loc[df_candidates.index[0]:df_candidates.index[-1]].shape[0]
        df.loc[df_candidates.index[0]:df_candidates.index[-1], "hyp_seq_id"] = df.loc[df_candidates.index[0]]["hyp_seq_id"]

    # TODO: should we save the index and restore it?
    df.reset_index(inplace=True)
    df.index = saved_index.values

    return df["hyp_sleep_candidate"], df["hyp_seq_id"], df["hyp_seq_length"]
